#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import base64
import datetime
import hmac
import json
import logging
import time
from hashlib import sha1

import requests
from munch import Munch

from api.model.Page import Page
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.execptions import FDExecuteException, NotSupportedFDException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext
from libs.utils import JSONEncoder

log = logging.getLogger('Ficus')


class SobeyHiveFactDatasourceContext(AbstractFactDatasourceContext):

    DEFAULT_CONTENT_TYPE = "application/json;charset=utf-8"
    # 数据对象字段深度，避免相互引用造成死循环
    FIELD_DEPT = 10

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        return 0

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        return False

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        return None

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def query(self, query: str, parameters: dict = None):
        """
        通过使用hive的全文检索来查询数据
        不支持 pageNum、pageSize、needCount参数
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        if not query:
            return None

        # 参数占位符处理
        query = self._fd_placeholder_replace(query, parameters)
        source_name = self.fd().get_source_name()
        try:
            json_query = json.loads(query)
        except Exception as e:
            error = f'事实库{source_name}执行query操作SQL错误：{query}, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)

        url = "http://"+self.fd().connection+"/sobeyhive-bp/v1/search?isCountPrivilege=false"

        # 构建http头信息, 签名5分钟失效,因此,可以只生成一次
        headers = self._build_http_headers('POST')

        try:
            r = requests.post(url, json=json_query, headers=headers)
            if r.ok and len(r.content) > 0:
                resource_search_result = Munch(r.json())
                if resource_search_result.queryResult:
                    query_result = resource_search_result.queryResult
                    page = Page(query_result.get('currentPage'), query_result.get('pageSize'), True, False)
                    page.set_result(query_result.get('result'), query_result.get('totalCount'))
                    return page
        except Exception as e:
            source_name = self.fd().get_source_name()
            error = f'事实库{source_name}执行query操作SQL：{json_query} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        return None

    def inserts(self, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        if not result_list:
            return
        if not isinstance(result_list, list):
            result_list = [result_list]

        url = f'http://{self.fd().connection}/sobeyhive-bp/v1/entity'

        # 构建http头信息, 签名5分钟失效,因此,可以只生成一次
        headers = self._build_http_headers('PATCH')

        i = 0
        while i < len(result_list):
            try:
                result = result_list[i]
                i += 1

                if isinstance(result, str):
                    result_str = result
                    result = json.loads(result)
                else:
                    result_str = json.dumps(result, cls=JSONEncoder)

                r = requests.patch(url, data=result_str, headers=headers)

                # 这里会返回结果，不会抛出异常
                if r.status_code == 500:
                    response = Munch(r.json())
                    code = response.code
                    # 入库验证不通过, B0005 新增验证不通过，B0192 部分更新验证不通过
                    if 'B0005' == code or 'B0192' == code:
                        if self._create_entity_obj_by_dict(result):
                            # 创建业务对象后重试一次,只有创建成功才重试，否则会导致死循环
                            i -= 1
                            continue
                elif not r.ok:
                    raise FDExecuteException(f"插入数据错误：{str(r.json())}")
            except Exception as e:
                source_name = self.fd().get_source_name()
                error = f'事实库{source_name}执行inserts操作发生异常, {str(e)}'
                log.error(error)
                raise FDExecuteException(error)

    def updates(self, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        self.inserts(result_list)

    def inserts_or_updates(self, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        self.inserts(result_list)

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def delete_all(self):
        """
        清空数据
        :return:
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :return:
        """
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def delete_conditions(self, condition_groups: list):
        raise NotSupportedFDException('sobeyhive不支持该操作')

    def _build_http_headers(self, request_method: str) -> dict:
        """
        构造请求头
        :param request_method:
        :return:
        """
        date_string = time.strftime("%a, %d %b %Y %H:%M:%S %z")
        signature = self._build_signature(request_method, date_string, 'SobeyHive')

        headers = {
            "Content-Type": self.DEFAULT_CONTENT_TYPE,
            # 这里使用credentials来设置站点有可能从ficus的A站点向hive的B站点中入，所以不能直接使用fd的site
            # 默认使用S1站点
            "sobeyhive-http-site": "S1" if self.fd().credentials is None else self.fd().credentials,
            "sobeyhive-http-system": "SobeyHive",
            "sobeyhive-http-system_authorization": f'SobeyHive SobeyHive:{signature}',
            "sobeyhive-http-date": date_string,
            "current-user-code": "admin"
        }
        return headers

    def _build_signature(self, request_method: str, date_string: str, s3secretkey: str):
        to_signature = f'{request_method}\n\n{self.DEFAULT_CONTENT_TYPE}\n{date_string}'
        hmac_code = hmac.new(s3secretkey.encode(), to_signature.encode(), sha1).digest()
        return base64.b64encode(hmac_code).decode()

    def _create_entity_obj_by_dict(self, entity: dict) -> bool:
        if 'entityData' not in entity:
            raise FDExecuteException(f'未设置entityData值，无法创建业务对象。{entity}')

        type_name = self.fd().target
        entity_data: dict = entity.get('entityData')

        # 1. 创建数据对象
        resource_type_id = self._create_resource_obj(type_name)
        if resource_type_id <= 0:
            return False

        # 2. 创建数据对象的字段
        self._create_resource_field_by_dict(resource_type_id, entity_data)

        # 3. 创建业务对象
        self._create_entity_obj(type_name, resource_type_id)

        return True

    def _create_resource_obj(self, type_name):
        """
        创建数据对象
        :param type_name:
        :return: success: id  fail: -1
        """
        headers = self._build_http_headers('POST')
        url = f'http://{self.fd().connection}/sobeyhive-bp/v1/resourceobj'

        # 构建请求参数
        request_params = {
            'resourceType': {
                'typeName': type_name
            }
        }

        r = requests.post(url, json=request_params, headers=headers)
        if r.ok and len(r.content) > 0:
            response_obj = r.json()
            if 'id' in response_obj:
                return int(response_obj.get('id'))

        return -1

    def _create_resource_field_by_dict(self, resource_type_id: int, entity_data: dict):
        """
        创建数据对象的字段
        :param resource_type_id:
        :param entity_data:
        :return:
        """
        if resource_type_id <= 0 or not entity_data:
            return

        for key, value in entity_data.items():
            field_name = str(key)

            # 跳过hive的内置字段
            if field_name.endswith('_'):
                continue
            if value is None:
                continue

            data_type = self._class2type(value)

            meta_model_field = dict()
            meta_model_field['fieldName'] = field_name

            if isinstance(value, dict):
                sub_resource_type_id = self._create_resource_obj(field_name)
                if resource_type_id <= 0:
                    continue

                self._create_resource_field_by_dict(sub_resource_type_id, value)
                meta_model_field['refModel'] = field_name
                data_type = 'object'
            elif isinstance(value, list):
                meta_model_field['multiple'] = True
                obj = value[0]
                if obj is None:
                    # TODO 这里就直接返回了吗？
                    continue

                if isinstance(obj, dict):
                    if obj is None:
                        return
                    sub_resource_type_id = self._create_resource_obj(field_name)
                    if resource_type_id <= 0:
                        continue

                    self._create_resource_field_by_dict(sub_resource_type_id, value)
                    meta_model_field['refModel'] = field_name
                    data_type = 'object'
                else:
                    data_type = self._class2type(obj)

            self._create_resource_field(resource_type_id, data_type, meta_model_field)

    def _class2type(self, obj) -> str:
        if isinstance(obj, bool):
            return 'boolean'
        elif isinstance(obj, bytes) or isinstance(obj, str):
            return 'string'
        elif isinstance(obj, float):
            return 'double'
        elif isinstance(obj, int):
            return 'long'
        elif isinstance(obj, datetime.date):
            return 'date'
        else:
            return 'string'

    def _create_entity_obj(self, type_name, resource_type_id):
        """
        创建业务对象
        :param type_name:
        :param resource_type_id:
        :return:
        """
        headers = self._build_http_headers('POST')
        url = f'http://{self.fd().connection}/sobeyhive-bp/v1/entityobj'

        # 构建请求参数
        request_params = {
            'entityTypeVO': {
                'typeName': type_name
            },
            'entityTypeAndResourceTypes': [{'mainData': 1, 'resourceTypeId': resource_type_id}]
        }

        try:
            r = requests.post(url, json=request_params, headers=headers)
            if not r.ok:
                raise FDExecuteException(f'数据源{self.fd().get_source_name()}创建业务对象失败,{str(r.content)}')
        except Exception as e:
            raise FDExecuteException(f'数据源{self.fd().get_source_name()}创建业务对象失败,{str(e)}')

    def _create_resource_field(self, resource_type_id, data_type, meta_model_field: dict):
        """
        创建数据对象字段
        :param resource_type_id:
        :param data_type:
        :param meta_model_field:
        :return:
        """
        headers = self._build_http_headers('POST')
        url = f'http://{self.fd().connection}/sobeyhive-bp/v1/resourceobj/field'
        # 构建请求参数
        request_params = {
            'resourceTypeId': resource_type_id,
            'fieldName': meta_model_field.get('fieldName')
        }

        try:
            if data_type:
                request_params['dataType'] = data_type
            if 'alias' in meta_model_field:
                request_params['alias'] = meta_model_field['alias']
            if 'description' in meta_model_field:
                request_params['description'] = meta_model_field['description']
            if 'minLength' in meta_model_field:
                request_params['minLen'] = meta_model_field['minLength']
            if 'maxLength' in meta_model_field:
                request_params['maxLen'] = meta_model_field['maxLength']
            if 'nullable' in meta_model_field:
                nullable = bool(meta_model_field['nullable'])
                must_input = 0 if nullable else 1
                request_params['mustInput'] = must_input
            if 'multiple' in meta_model_field:
                multiple = bool(meta_model_field['multiple'])
                is_array = 1 if multiple else 0
                request_params['isArray'] = is_array
            if 'refModel' in meta_model_field:
                request_params['refResourceTypeName'] = meta_model_field['refModel']

            r = requests.post(url, json=request_params, headers=headers)
            if not r.ok:
                raise FDExecuteException(f'数据源{self.fd().get_source_name()}创建数据对象字段失败,对象字段值：{str(meta_model_field)},{str(e)}')
        except Exception as e:
            raise FDExecuteException(f'数据源{self.fd().get_source_name()}创建数据对象字段失败,对象字段值：{str(meta_model_field)},{str(e)}')











#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from enum import Enum

from redis import StrictRedis

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.execptions import DatasourceNotFoundException, FDExecuteException
from factdatasource.dao.FactDatasource import DatasourceListener, MultipleBaseDatasource, BaseDatasource
from libs.utils import Singleton


class RedisTypeEnum(Enum):
    SINGLE = 'SINGLE '  # 单节点
    CLUSTER = 'CLUSTER'  # 集群
    SENTINEL = 'SENTINEL'  # 哨兵


class MultipleRedisDatasource(DatasourceListener, MultipleBaseDatasource, Singleton):
    """
    管理fd中的所有REDIS类型的数据源
    """

    def __init__(self):
        # source_name --> RedisDatasource
        self._target_dataSources = dict()

    def get_datasource_type(self):
        """
        获取数据源类型
        :return:
        """
        return FactDatasourceTypeEnum.REDIS

    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'添加数据源参数错误:source_name、url都不能为空')

        for target in self._target_dataSources.values():
            if target.url == url and target.credentials == credentials:
                # url 和 credentials 都相同，说明是同一个数据库连接就不再重复创建数据源了
                target.add_source_name(source_name)
                self._target_dataSources[source_name] = target
                return

        # 开始创建数据源
        redis_data_source = RedisDatasource(source_name, url, credentials)
        redis_data_source.start()
        self._target_dataSources[source_name] = redis_data_source

    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'删除数据源参数错误:source_name不能为空')

        self.delete_datasource_type(source_name)
        self.add_datasource_type(source_name, url, credentials)

    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'删除数据源参数错误:source_name不能为空')

        target: RedisDatasource = self._target_dataSources.pop(source_name)
        if target:
            if target.only_one_source():
                target.close_client()
            else:
                target.remove_source_name(source_name)

    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """
        source_name = self.determine_current_lookup_key()
        if not source_name:
            raise FDExecuteException('未设置操作源,无法获取数据源信息。')

        if source_name in self._target_dataSources.keys():
            return self._target_dataSources[source_name]
        raise DatasourceNotFoundException('未发现数据源%s。' % source_name)

    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """
        target: RedisDatasource = self.get_data_source()
        return target.get_client()

    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """
        target: RedisDatasource = self.get_data_source()
        return target.close_client()


class RedisDatasource(BaseDatasource):

    def __init__(self, source_name, url, credentials):
        self.init(source_name, url, credentials)
        self.client: StrictRedis = None

    def start(self):
        """
        TODO 连接池暂时就用默认的连接池
        :return:
        """
        conn_url = self._parse_url()
        type = conn_url[0]

        kwargs = {
            "decode_responses": True
        }

        if type == RedisTypeEnum.CLUSTER:
            from rediscluster import StrictRedisCluster
            self.client = StrictRedisCluster(startup_nodes=conn_url[1], **kwargs)
        elif type == RedisTypeEnum.SENTINEL:
            from redis.sentinel import Sentinel
            sentinel = Sentinel(conn_url[2], **kwargs)
            self.client = sentinel.master_for(conn_url[1])
        else:
            self.client = StrictRedis.from_url(conn_url[1], **kwargs)

    def get_client(self):
        return self.client

    def close_client(self):
        if self.client:
            self.client.connection_pool.disconnect()

    def _parse_url(self):
        """
        根据配置将url转换为可以直接连接的url
        REIDS原始配置如下:
        单机：ip:port
        集群（英文逗号分割）： ip:port,ip:port,ip:port
        哨兵（第一个表示主节点名字，英文分号分割）：MASTERNAME;哨兵IP1:PORT1;哨兵IP2:PORT2;哨兵IP3:PORT3
        :return:
        """

        # 获取密码
        password = None
        if self.credentials:
            if ':' in self.credentials:
                password = str(self.credentials).split(':', 1)[1]
            else:
                password = self.credentials

        result = []

        if ',' in self.url:
            # 集群配置
            type = RedisTypeEnum.CLUSTER
            for url_split in str(self.url).split(','):
                url_tmp = {}
                if ':' in url_split:
                    url_tmp['host'] = url_split.split(':', 1)[0]
                    url_tmp['port'] = int(url_split.split(':', 1)[1])
                else:
                    url_tmp['host'] = url_split
                if password:
                    url_tmp['password'] = password

                result.append(url_tmp)
            return (type, result, )
        elif ';' in self.url:
            # 哨兵配置  [('localhost', 26379)]
            type = RedisTypeEnum.SENTINEL
            url_split = str(self.url).split(';')
            if len(url_split) < 2:
                raise IllegalArgumentException(f'redis连接信息配置错误，无法进行连接：{self.url}')
            # 第一个是主节点名字
            master_service = url_split[0]

            for host_port in url_split[1:]:
                if ':' in host_port:
                    result.append((host_port.split(':', 1)[0], host_port.split(':', 1)[1]))
                else:
                    result.append((host_port,))
            return (type, master_service, result, )
        else:
            # 普通单节点
            type = RedisTypeEnum.SINGLE
            #构造url:  redis://[:password]@localhost:6379/0
            password = ':'+str(password)+'@' if password else ''
            url = f'redis://{password}{self.url}'
            return (type, url)





from api.handler.outputer.IBaseOutputer import IBaseOutputer


class ISimpleOutputer(IBaseOutputer):

    def send_output_result(self, code, result_list, output_fd_codes):
        """
        发送任务到FD
        :param code:
        :param result_list:
        :param output_fd_codes:
        :return:
        """

        insert_cache = {}
        update_cache = {}
        upsert_cache = {}

        # 把任务放入缓存
        for serializable_output_wrapper in result_list:
            output_fd = self.find_output_fd(output_fd_codes, serializable_output_wrapper.index())
            self.put_in_cache(code, insert_cache, update_cache, upsert_cache, output_fd, serializable_output_wrapper)

        # 清理缓存中的任务
        self.flush_cache(code, insert_cache, update_cache, upsert_cache)



"""
使用标准的 AES/CBC/PKCS5PADDING  PBKDF2WithHmacSHA256  来加解密字符串
@author sun
@Date 2019-11-30 11:27
@version 1.0
"""

import base64
import hashlib

from Crypto import Random
from Crypto.Cipher import AES

block_size = 16
pad = lambda s: s + (block_size - len(s) % block_size) * chr(block_size - len(s) % block_size)
unpad = lambda s: s[0:-ord(s[-1:])]
iv = Random.new().read(AES.block_size) # Random IV

def __get_private_key(secret_key, salt):
    return hashlib.pbkdf2_hmac('SHA256', secret_key.encode(), salt.encode(), 65536, 32)

def encrypt_with_AES(message, secret_key, salt):
    """
    AES加密
    :param message: 明文. 只能是英文数字符号
    :param secret_key:
    :param salt:
    :return:
    """
    private_key = __get_private_key(secret_key, salt)
    message = pad(message)
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    cipher_bytes = base64.b64encode(iv + cipher.encrypt(message))
    return bytes.decode(cipher_bytes)

def decrypt_with_AES(encoded, secret_key, salt):
    """
    AES解密
    :param encoded:
    :param secret_key:
    :param salt:
    :return:
    """
    private_key = __get_private_key(secret_key, salt)
    cipher_text = base64.b64decode(encoded)
    iv = cipher_text[:AES.block_size]
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    plain_bytes = unpad(cipher.decrypt(cipher_text[block_size:]))
    return bytes.decode(plain_bytes)


if __name__ == '__main__':
    # 加解密测试

    secret_key = "$0bEyHive&2o1Six"
    salt = "JXTYp9icQaTzs4"
    plain_text = "English@12_d:a&^%*"

    cipher = encrypt_with_AES(plain_text, secret_key, salt)
    print("Cipher: " + cipher)

    decrypted = decrypt_with_AES(cipher, secret_key, salt)
    print("Decrypted " + decrypted)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.execptions import NotSupportedFDException


def get_multiple_datesource(fd_type: FactDatasourceTypeEnum):
    if fd_type == FactDatasourceTypeEnum.JDBC:
        from factdatasource.dao.jdbc.MultipleJdbcDatasource import MultipleJdbcDatasource
        return MultipleJdbcDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.ES:
        from factdatasource.dao.es.MultipleEsDatasource import MultipleEsDatasource
        return MultipleEsDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.MONGO:
        from factdatasource.dao.mongo.MultipleMongoDatasource import MultipleMongoDatasource
        return MultipleMongoDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.REDIS:
        from factdatasource.dao.redis.MultipleRedisDatasource import MultipleRedisDatasource
        return MultipleRedisDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.KAFKA:
        from factdatasource.dao.kafka.MultipleKafkaDatasource import MultipleKafkaDatasource
        return MultipleKafkaDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.GRAPH:
        from factdatasource.dao.jdbc.MultipleJdbcDatasource import MultipleJdbcDatasource
        return MultipleJdbcDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.AMQP:
        from factdatasource.dao.amqp.MultipleAmqpDatasource import MultipleAmqpDatasource
        return MultipleAmqpDatasource.instance()
    elif fd_type == FactDatasourceTypeEnum.JMS:
        from factdatasource.dao.stomp.MultipleStompDatasource import MultipleStompDatasource
        return MultipleStompDatasource.instance()
    else:
        raise NotSupportedFDException(f'暂未支持{fd_type}类型的数据源')



"""
注解类,用于表名Taskhandler的注解的,TODO 这里肯定是错的
"""

TASK_HANDLERS = dict()


def TaskHandler(name):
    """
    TaskHandler的类注解
    :param name:
    :return:
    """

    def decorate(cls):
        TASK_HANDLERS[name] = cls()
        return cls      # 让装饰器最后还是返回cls. 否则经过装饰器包装的类的返回时None

    return decorate



from enum import Enum


class TransformSourceTypeEnum(Enum):
    FIELD = "FIELD",  # 字段, 如果是这种, 那么在FD最终写入数据库的时候, 会再通过这个转换器转换一次结果
    MAPPING = "MAPPING"  # 映射, 如果是这种, 那么在映射转换过程中, 会通过这个转换器转换一次结果, 不填就是完整拷贝源的值到目标, 如果源的值是多个, 根据目标字段类型决定默认值是什么, 字符串默认为空格拼接 数字是相加 布尔是且 日期取最大


def init_transform_by_field(id):
    from factdatasource.dao.jdbc.JdbcDatasourceDao import JdbcDatasourceDao
    dao = JdbcDatasourceDao.instance()
    sql = f"""SELECT id,sourceType,sourceId,transformer,params,orders,updateTime FROM `sc_metamodel_transform` WHERE id=:id and sourceType = 'FIELD'"""
    code_dict = {'id': id}
    rows = dao.select_all(sql, code_dict)

    if rows is not None and len(rows) > 0:
        return [MetaModelTransform(model["id"], model["sourceType"], model["sourceId"], model["transformer"],
                                   model["params"],
                                   model["orders"], model["updateTime"]) for model in rows]

    return None


class MetaModelTransform:

    def __init__(self, id=None, sourceType=None, sourceId=None, transformer=None, params=None, orders=None,
                 updateTime=None):
        self.id = id
        self.sourceType: TransformSourceTypeEnum = sourceType
        self.sourceId = sourceId
        self.transformer = transformer
        self.orders = orders
        self.updateTime = updateTime
        if params:
            import json
            self.params = json.loads(params)

    def __str__(self) -> str:
        return f"{self.transformer}_{self.sourceType}_{self.sourceId}_{self.id}"

    def timeString(self) -> str:
        if self.updateTime is None:
            import datetime
            return datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
        else:
            return self.updateTime.strftime("%Y%m%d%H%M%S%f")



import threading
from abc import abstractmethod
from datetime import datetime
from queue import Queue
from concurrent.futures import ThreadPoolExecutor, as_completed

from munch import Munch

from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.outputer.IBatchOutputer import IBatchOutputer
from api.model.BatchOutputPipe import BatchOutputPipe
from api.model.ResultVO import *
from client import DataCrawlClient
from schedule.utils.log import TaskLogFileAppender


class AbstractBatchCrawl(ITaskHandler, IBatchOutputer, ICacheAbleHandler):
    # 由于这个的实现有可能是 单例的,所以要使用ThreadLocal
    def __init__(self):
        super().__init__()
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__execution_message_local = threading.local()
        self.__task_log_id_local = threading.local()
        self.__tasks_local_host = threading.local()

        self.killed = False

    def __action(self, data_crawl, output_stream, params, is_finished, messages):
        # 异步线程来对数据进行输入
        from schedule import ShardContext
        try:
            self.set_local_code(data_crawl.site + "_" + data_crawl.projectCode + "_" + data_crawl.code)
            self.set_process_id(params.get("__processLogId__"))
            self.__task_log_id_local.key = params.get("__logId__")
            TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                               params["__logId__"])
            if "__sharding_index__" in params and "__sharding_total__" in params:
                from schedule.ShardContext import Sharding
                ShardContext.set_sharding(Sharding(params["__sharding_index__"], params["__sharding_total__"]))
            self.__execution_message_local.content = []
            self.do_crawl(BatchOutputPipe(output_stream, len(data_crawl.outputFdCodes)), params)
        finally:
            is_finished[0] = True
            messages[0] = "success" if self.__execution_message_local.content is None or len(
                self.__execution_message_local.content) == 0 else str(self.__execution_message_local.content)
            # 清理 MessageLocal 和 LocalCode
            self.clear_local_code()
            self.clear_process_id()
            self.__task_log_id_local.key = None
            self.__execution_message_local.content = None
            ShardContext.reset()
            TaskLogFileAppender.end_log()

    def execute(self, params: dict):

        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在crawl的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有Crawl的信息")

        self.__tasks_local_host.tasks = []

        output_stream = Queue()

        self.killed = False

        with ThreadPoolExecutor(max_workers=2, thread_name_prefix="batch-ce-") as executor:
            # 这个的逻辑是这样的:
            # 1.传入一个 队列进去,然后异步等待这个线程做完
            try:
                # 找到对应的dataCrawl
                data_crawl: Munch = DataCrawlClient.get(params["site_"], params["projectCode_"], params["code_"])

                if data_crawl is None or data_crawl.type != "CUSTOM":
                    return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['code_']}的Crawl,或者该Crawl类型不为Custom")

                is_finished = [False]
                messages = [None]

                self.__tasks_local_host.tasks.append(
                    executor.submit(self.__action, data_crawl, output_stream, params, is_finished, messages))

                # 增加发送结果的线程
                self.__tasks_local_host.tasks.append(
                    executor.submit(self.batch_output, params["code_"], data_crawl.outputFdCodes,
                                    is_finished, output_stream,params))

                # 阻塞主线程
                for future in as_completed(self.__tasks_local_host.tasks):
                    try:
                        data = future.result()
                    except Exception as e:
                        import traceback
                        return ResultVO(FAIL_CODE, f"执行失败,Code:{params['code_']}的Crawl,原因:\n{traceback.format_exc()}")

                if messages[0] is not None:
                    return ResultVO(SUCCESS_CODE, messages[0])

                return SUCCESS
            finally:
                self.__tasks_local_host.tasks.clear()
                executor.shutdown(wait=True)
                output_stream.queue.clear()

    def kill(self):
        self.killed = True
        if self.__tasks_local_host.tasks is not None:
            for task in self.__tasks_local_host.tasks:
                try:
                    task._stop()
                except:
                    pass

    def is_killed(self):
        return self.killed

    def get_execution_message_cache(self):
        """
        返回message_cache
        :return:
        """
        return self.__execution_message_local.content

    def get_batch_size(self) -> int:
        # TODO 先写死100
        return 100

    def get_code_thread_local(self):
        """
        实现上下文的code
        :return:
        """
        return self.__code_local_host

    def get_process_thread_local(self):
        """
        上下文的Id
        :return:
        """
        return self.__process_id_local

    def get_task_log_id(self):
        return self.__task_log_id_local

    @abstractmethod
    def do_crawl(self, output_stream: BatchOutputPipe, params: dict):
        """
        真正执行数据挖掘的逻辑
        :param output_stream: 数据的输出
        :param params: 需要的参数
        :return:
        """



# -*- coding: utf-8 -*-
import socket
import re
import base64
import gzip

try:
    import urllib.request as urllib2
    from urllib.error import HTTPError
    from urllib.error import URLError
except ImportError:
    import urllib2
    from urllib2 import HTTPError
    from urllib2 import URLError
    from StringIO import StringIO

_URL_REGEX = re.compile(
    r'^(?:http)s?://'  # http:// or https://
    r'(([A-Z0-9_~!.%]+):([A-Z0-9_~!.%]+)@)?'  # basic authentication -> username:password@
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?)|'  # domain name without `.`
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)


def get_url_and_basic_auth(addr_url):
    addr = addr_url

    match_obj = _URL_REGEX.match(addr)
    groups = match_obj.groups()
    if(groups[0] is not None):
        addr = addr.replace(groups[0], "")
        user_name = groups[1]
        user_psw = groups[2]
        ori_auth = ("%s:%s" % (user_name, user_psw)).encode()
        auth_str = base64.standard_b64encode(ori_auth).decode()
        return (addr, auth_str)
    else:
        return (addr, None)


class Request(urllib2.Request, object):

    def __init__(self, url, data=None, headers={},
                 origin_req_host=None, unverifiable=False,
                 method=None):
        url_match = _URL_REGEX.match(url)
        if url_match is None:
            raise URLError("Unvalid URL")
        url_obj = get_url_and_basic_auth(url)
        url_addr = url_obj[0]
        url_auth = url_obj[1]
        try:
            super(Request, self).__init__(url_addr, data=data, headers=headers,
                                          origin_req_host=origin_req_host, unverifiable=unverifiable,
                                          method=method)
        except TypeError:
            super(Request, self).__init__(url_addr, data=data, headers=headers,
                                          origin_req_host=origin_req_host, unverifiable=unverifiable)
            self.get_method = lambda: method if method is not None else "GET"
        if url_auth is not None:
            self.add_header('Authorization', 'Basic %s' % url_auth)


def urlopen(url, data=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
            cafile=None, capath=None, cadefault=False, context=None):
    if isinstance(url, urllib2.Request):
        request = url
    elif isinstance(url, str):
        request = Request(url)
    else:
        raise URLError("Unvalid URL")
    request.add_header("Accept-encoding", "gzip")
    res = urllib2.urlopen(request, data=data, timeout=timeout,
                          cafile=cafile, capath=capath, cadefault=cadefault, context=context)

    if res.info().get("Content-Encoding") == "gzip":
        try:
            # python2
            f = gzip.GzipFile(fileobj=StringIO(res.read()))
        except NameError:
            f = gzip.GzipFile(fileobj=res)
    else:
        f = res

    return f



"""
任务记录相关的操作
"""
from datetime import datetime

import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def update_task_status_to_execute(log_id, instance_address):
    """
    更新任务的状态到正在执行
    :param log_id:
    :param instance_address:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sjtl-service/{log_id}/execute", params={'instanceAddress': instance_address},
                       method="post")
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def update_task_status_to_finished(log_id, execute_result, triggerProcess):
    """
    更新任务的状态到结束
    :param log_id:
    :param execute_result:
    :return:
    """
    if execute_result is None:
        return

    check_instance_avaliable()

    request = None
    if not isinstance(execute_result, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(execute_result, dict):
            # 如果是dict类型的,直接发
            request = execute_result
    else:
        # 说明是munch的,那么就转成Dict的
        request = execute_result.toDict()

    try:
        r = do_service(f"remote/sjtl-service/{log_id}/finished?triggerProcess={triggerProcess}", data=request,
                       method="post")
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def count_failed_task_log_by_retry_log_id(retry_log_id):
    """
    根据retryLogId查询失败的任务数量
    :param retry_log_id:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sjtl-service/retry/{retry_log_id}")
        if r is not None:
            return r
        else:
            return 0
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def count_success_task_log_by_job(job_id, update_time):
    """
    统计执行成功的任务的数量
    :param job_id:
    :param update_time: 更新时间, 格式: yyyy-MM-dd HH:mm:ss 的字符串 或者 datetime 类型
    :return:
    """
    check_instance_avaliable()

    request = None
    if isinstance(update_time, str):
        request = update_time
    elif isinstance(update_time, datetime):
        request = update_time.strftime("%Y-%m-%d %H:%M:%S")

    try:
        r = do_service(f"remote/sjtl-service/{job_id}", params={'updateTime': request})
        if r is not None:
            return r
        else:
            return 0
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def get_task_log_by_id(log_id):
    """
    根据logId查询任务详情
    :param log_id:
    :return:
    """
    check_instance_avaliable()
    try:
        r = do_service(f"remote/sjtl-service/task/{log_id}")
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def doing_tasks(project_code):
    """
    查询正在执行的任务
    :param project_code:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sjtl-service/{project_code}/doing")
        if r is not None:
            if isinstance(r, list):
                return [Munch(x) for x in r]
            else:
                return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def update_task_progress(log_id, progress):
    """
    更新任务的进度
    :param log_id:
    :param progress:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sjtl-service/{log_id}/progress", params={"progress": progress})
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



REMOTE_YML_CONFIG = {}


def Value(value):
    """
    spring-cloud的配置获取注解
    :param value: ${zookeeper.timeout:10000}   冒号后面是默认值 可以没有
    :return:
    """

    def decorate(func):
        def wrapper(*arg, **kvargs):  # 包装的具体过程
            # 这里访问eureka去读取配置文件
            return get_value(value)

        return wrapper

    return decorate


def __get_value(property: str):
    """
        从配置中获取key对应的value
        需要注意有递归的情况
        :param value:
        :return:
        """
    tup = _do_parse_expression(property)

    if tup[0] == "":
        # 说明不满足 ${xxxx:xx} 这种格式.  直接返回值
        return _convertStr2Other(tup[1])

    # 说明满足占位符的格式,开始解析占位符
    # arr[0]有值,说明需要到REMOTE_YML_CONFIG里面去找
    keys = tup[0].split(".")
    tmp = REMOTE_YML_CONFIG
    try:
        for key in keys:
            tmp = tmp[key]
        # 这个地方的tmp 有可能还需要递归做处理
        return get_value(tmp)
    except Exception as e:
        # 默认值
        if tup[1] is None or tup[1] == ":" or tup[1] == "":
            return None
        else:
            return _convertStr2Other(tup[1][1:])


def get_value(property: str):
    """
    从配置中获取key对应的value
    需要注意有递归的情况
    :param value:
    :return:
    """
    # 这里其实还需要根据${}正则的方式,把字符串拆开. 比如: ${sobeycs.redis.host}:${sobeycs.redis.port} 其实应该拆成3个,分别的递归处理.最后再合并

    if not isinstance(property, str):
        return property

    import re
    l = re.compile("(\${[\w\.-]+[:[\w]+]*})").split(property)

    l = list(filter(lambda x: x if x is not None and x!="" else "",l))

    if len(l) >1 :
        return "".join(map(lambda x:str(x),map(__get_value, l)))
    elif len(l) == 1:
        return __get_value(l[0])
    else:
        return property


def _do_parse_expression(value):
    """
    解析表达式
    :param value:
    :return: 数组, [0]是在remote上的key  [1]是默认值
    """
    if not isinstance(value, str) or not value.startswith("${"):
        return "", value

    import re
    aaa = re.findall(r"\${((?:\w|\.|-)*)(:(\w|/)*)*}", value)
    return aaa[0]


def _convertStr2Other(value):
    """
    数据转换,把字符串的结果进行类型转换尝试,转换成为 int float boolean
    :param value:
    :return:
    """

    if value is None:
        return None

    if isinstance(value, bool) or isinstance(value, int) or isinstance(value, float):
        return value

    # 先尝试bool
    if "true" == value.lower():
        return True
    elif "false" == value.lower():
        return False

    try:
        # 尝试转换成为int
        return int(value)
    except ValueError:
        try:
            # 尝试转换成为float
            return float(value)
        except ValueError:
            # 都不是,那么就是字符串
            return value


def deep_search_and_merge(dict1, dict2):
    """
    深度遍历以及合并两个 dict 类型的数据. 要求相同key的值进行合并,如果值是dict,那么递归合并
    :param dict1:
    :param dict2:
    :return:
    """
    if dict2 is None or not isinstance(dict2, dict):
        return

    for key in dict2.keys():
        if key not in dict1.keys():
            dict1[key] = dict2[key]
        else:
            val1 = dict1[key]
            val2 = dict2[key]
            if val1 is None or val2 is None:
                # 如果两个中有一个是None,那么就不用递归了,直接替换
                dict1[key] = val2
            elif isinstance(val1, dict) and isinstance(val2, dict):
                # 只有两个都是字典类型的情况下,再递归调用
                deep_search_and_merge(dict1[key], dict2[key])
            else:
                # 说明dict1或dict2中某一个相同的key的value是基础类型,那么就直接替换
                dict1[key] = val2



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging

from munch import Munch

from api.model.GraphNode import GraphNode
from api.model.GraphRelation import GraphRelation
from api.model.Page import Page
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.dao.graph.GraphDatasourceDao import GraphDatasourceDao
from factdatasource.execptions import FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext
from factdatasource.persistence.SqlableDeleteCondition import SqlableDeleteCondition

log = logging.getLogger('Ficus')


class GraphdbFactDatasourceContext(AbstractFactDatasourceContext, SqlableDeleteCondition):
    DEFAULT_GROUP = 'ficus'

    @property
    def dao(self) -> GraphDatasourceDao:
        return GraphDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        return 0

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        return False

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        return None

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        return None

    def query(self, query: str, parameters: dict):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        source_name = self.fd().get_source_name()

        page = None
        if parameters and 'pageNum_' in parameters.keys() and 'pageSize_' in parameters.keys():
            # 需要分页的
            param = {
                'pageNum': parameters.get('pageNum_'),
                'pageSize': parameters.get('pageSize_'),
                'needCount': parameters.get('needCount_') if parameters.get('needCount_') else False
            }
            page = Page(**param)

        # 参数占位符处理
        query = self._fd_placeholder_replace(query, parameters)

        sql = self._query_page_sql(query, page)

        log.debug(f'事实库{source_name}执行query操作SQL：{sql}')
        try:
            costomer_context.set_source(source_name)

            result = self.dao.query(sql, parameters)
            if page:
                total = None
                if page.need_count():
                    total = self.dao.query_total(query, parameters)
                page.set_result(result, total)
                return page
            else:
                return Page.no_page(result)
        except Exception as e:
            error = f'事实库{source_name}执行query操作SQL：{sql} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def _query_page_sql(self, query: str, page: Page = None):
        """
        构造分页语句
        :param query: 原始SQL
        :param page:
        :return:
        """
        if page:
            sql = f'{query} LIMIT {page.start_row},{page.end_row}'
        else:
            sql = query
        return sql

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                if not isinstance(result, GraphNode) and not isinstance(result, GraphRelation):
                    # 不是图结构就无法保存
                    log.warning(f'保存图结构数据类型错误，放弃保存数据:{result}')
                    rrr.append(Munch({"success": False, "error": "节点并不是图节点,无法保存", "content": result}))
                    continue
                self._inner_insert(table, result)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def _inner_insert(self, table: str, result):
        try:
            if isinstance(result, GraphNode):
                # 插入节点
                self._inner_insert_node(table, result)
            elif isinstance(result, GraphRelation):
                # 插入关系
                self._inner_insert_relation(table, result)
        except Exception as e:
            source_name = self.fd().get_source_name()
            error = f'事实库{source_name}执行inserts操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)

    def _inner_insert_node(self, table: str, graph_node: GraphNode):
        graph_node['GROUP'] = self.DEFAULT_GROUP
        self.dao.create_node(table, graph_node)

    def _inner_insert_relation(self, table: str, graph_relation: GraphRelation):
        graph_relation['GROUP'] = self.DEFAULT_GROUP
        self.dao.create_relation(table, graph_relation)

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                if not isinstance(result, GraphNode) and not isinstance(result, GraphRelation):
                    # 不是图结构就无法修改
                    rrr.append(Munch({"success": False, "error": "节点并不是图节点,无法保存", "content": result}))
                    continue
                self._inner_update(table, result)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def _inner_update(self, table: str, result) -> bool:
        try:
            if isinstance(result, GraphNode):
                # 修改节点
                return self._inner_update_node(table, result)
            elif isinstance(result, GraphRelation):
                # 修改关系
                return self._inner_update_relation(table, result)
        except Exception as e:
            source_name = self.fd().get_source_name()
            error = f'事实库{source_name}执行inserts操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)

    def _inner_update_node(self, table, graph_node: GraphNode) -> bool:
        if graph_node.guid and graph_node.tags:
            return self.dao.update_node_tags(table, self.DEFAULT_GROUP, graph_node)
        else:
            # 信息不全无法进行修改
            log.warning(f'修改图节点时信息不全无法修改{graph_node}')
            return False

    def _inner_update_relation(self, table, graph_relation: GraphRelation) -> bool:
        if graph_relation.guid and graph_relation.tags:
            return self.dao.update_relation_tags(table, self.DEFAULT_GROUP, graph_relation)
        else:
            # 信息不全无法进行修改
            log.warning(f'修改图节点关系时信息不全无法修改{graph_relation}')
            return False

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                if not isinstance(result, GraphNode) and not isinstance(result, GraphRelation):
                    # 不是图结构就无法保存
                    rrr.append(Munch({"success": False,"error":"节点并不是图节点,无法保存","content":result}))
                    continue
                if not self._inner_update(table, result):
                    # 没有修改成功，尝试插入操作
                    try:
                        self._inner_insert(table, result)
                    except Exception as e:
                        # 重新尝试一次修改
                        self._inner_update(table, result)
                rrr.append(Munch({"success": True}))
        finally:
            costomer_context.clear_source()
        return rrr

    def delete_all(self):
        """
        清空数据
        :return:
        """
        table = self.fd().target
        source_name = self.fd().get_source_name()
        try:
            costomer_context.set_source(source_name)
            delete_relation_sql = f'DELETE RELATION FROM {table}'
            self.dao.delete_by_query(delete_relation_sql)
            delete_node_sql = f'DELETE NODE FROM {table}'
            self.dao.delete_by_query(delete_node_sql)
        except Exception as e:
            error = f'事实库{source_name}执行delete_all操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :return:
        """
        source_name = self.fd().get_source_name()
        try:
            costomer_context.set_source(source_name)
            self.dao.delete_by_query(query)
        except Exception as e:
            error = f'事实库{source_name}执行delete操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete_conditions(self, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param condition_groups:
        :return:
        """
        source_name = self.fd().get_source_name()
        try:
            costomer_context.set_source(source_name)
            query = self.delete_condition_sql(self.fd().target, condition_groups)
            self.dao.delete_by_query(query.replace("DELETE FROM", "DELETE NODE FROM", 1))
        except Exception as e:
            error = f'事实库{source_name}执行delete操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from factdatasource.persistence.jdbc.dialect.MysqlJdbcFactDatasourceContext import MysqlJdbcFactDatasourceContext


class VernoxJdbcFactDatasourceContext(MysqlJdbcFactDatasourceContext):
    pass


from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer


class ChineseToNumberTransformer(ITransformer):
    """
    中文转数字
    """

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):
        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # 能进来的只能是字符串
        old_value = data[target_model_field.fieldName]

        if old_value is None:
            return {target_model_field.fieldName: None}

        import cn2an
        return {target_model_field.fieldName: cn2an.cn2an(old_value)}



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22


class FactDatasourceException(Exception):
    pass


class CustomerDaoContextNotFoundException(FactDatasourceException):
    pass


class DatasourceNotFoundException(FactDatasourceException):
    pass


class FDExecuteException(FactDatasourceException):
    pass


class NotSupportedFDException(FactDatasourceException):
    pass


class FDNotExistsException(FactDatasourceException):
    pass


class FDConfigException(FactDatasourceException):
    pass


class FDOperationException(FactDatasourceException):
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/23
from elasticsearch import Elasticsearch

from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton


class EsDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.ES)
        return data_source

    @property
    def client(self) -> Elasticsearch:
        return self.datasource.get_client()

    def count(self, index=None, doc_type=None, body=None, **params):
        """
        获取数据条数
        :param index:
        :param doc_type:
        :param body:
        :param params:
        :return:
        """
        count_result = self.client.count(index=index, doc_type=doc_type, body=body, params=params)
        if count_result:
            return count_result.get('count')

    def simple_query(self, index=None, doc_type=None, body=None, **params):
        """
        简单查询，直接返回数据结果
        :param index:
        :param doc_type:
        :param body:
        :param params:
        :return:
        """
        result = self.client.search(index=index, doc_type=doc_type, body=body, params=params)
        hits = result['hits']['hits']
        if hits:
            return [source['_source'] for source in hits]

    def query(self, index=None, doc_type=None, body=None, **params):
        """
        普通查询
        :param index:
        :param doc_type:
        :param body:
        :param params:
        :return:
        """
        return self.client.search(index=index, doc_type=doc_type, body=body, params=params)

    def delete_by_query(self, index, body, doc_type=None, **params):
        """
        根据查询结果进行删除，es版本需要5.0以上才支持
        :param index:
        :param body:
        :param doc_type:
        :param params:
        :return:
        """
        params['refresh'] = 'true'
        return self.client.delete_by_query(index=index, doc_type=doc_type, body=body, params=params)

    def create(self, index, doc_type, id, body, **params):
        """
        添加数据
        :param index:
        :param doc_type:
        :param id:
        :param body:
        :param params:
        :return:
        """
        return self.client.create(index=index, doc_type=doc_type, id=id, body=body, params=params)

    def partial_update(self, index, doc_type, id, body=None, **params):
        """
        部分更新
        :param index:
        :param doc_type:
        :param id:
        :param body:
        :param params:
        :return:
        """
        upbody = {
            'doc': body
        }
        return self.client.update(index, doc_type, id, body=upbody, params=params)

    def upsert(self, index, doc_type, body, id=None, **params):
        """
        修改或添加
        :param index:
        :param doc_type:
        :param body:
        :param id:
        :param params:
        :return:
        """
        return self.client.index(index=index, doc_type=doc_type, body=body, id=id, params=params)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/28
import json

from libs.utils import JSONEncoder


class Page(object):

    def __init__(self, pageNum=1, pageSize=20, needCount=False, reasonable=True):
        """
        分页查询结果对象
        :param pageNum: 页数
        :param pageSize: 分页大小
        :param needCount: 需不需要进行总数查询
        :param reasonable: 分页合理化
        """
        if reasonable:
            self.pageNum = pageNum if pageNum > 0 else 1
            self.pageSize = pageSize if pageSize > 0 else 20
        else:
            self.pageNum = pageNum
            self.pageSize = pageSize
        # 进不进行count查询
        self.needCount = needCount
        self.result = None
        self.total = None

    @property
    def start_row(self):
        return (self.pageNum-1)*self.pageSize

    @property
    def end_row(self):
        return self.start_row + self.pageSize*self.pageNum

    def need_count(self):
        return self.needCount

    def set_result(self, result: list, total: int):
        self.result = result
        self.total = total

    def to_json(self):
        return {
            'pageNum': self.pageNum,
            'pageSize': self.pageSize,
            'needCount': self.needCount,
            'total': self.total,
            'result': self.result
        }

    @staticmethod
    def no_page(result: list):
        """
        不进行分页时，返回的page对象
        :param result:
        :return:
        """
        page = Page(0, 0, False, False)
        page.set_result(result, -1)
        return page

    def __repr__(self):
        return json.dumps(self.to_json(), cls=JSONEncoder)




from schedule.utils.log.TaskLogger import TaskLogger


class ILogAbleHandler:
    def __init__(self):
        self._task_logger = None

    @property
    def task_logger(self)->TaskLogger:
        "日志记录方法,有两个方法: log(message) 以及 error(exception)"
        return self._task_logger or TaskLogger(None)

    @task_logger.setter
    def task_logger(self, value:TaskLogger):
        self._task_logger = value


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/31
from unittest import TestCase

from factdatasource.dao.jdbc.MultipleJdbcDatasource import MultipleJdbcDatasource
from service.FactDatasourceService import FactDatasourceService


def register_default_datasource():
    # ficus数据库配置
    credentials = 'sobeyhive:$0bEyHive&2o1Six'
    url = 'jdbc:mysql://localhost:3306/ficus'
    datasource = MultipleJdbcDatasource.instance()
    datasource.add_datasource_type(datasource.DEFAULT_DATASOURCE, url, credentials)


class TestFactDatasourceService(TestCase):

    def test_get_fd_by_code(self):
        register_default_datasource()
        fd_service = FactDatasourceService.instance()
        fd = fd_service.get_fd_by_code('test_source_1')
        print(fd.get_source_name())

        fdx = fd_service.get_fd_by_code('test_source_x')
        print(fdx)

    def test_exists_fds(self):
        register_default_datasource()
        fd_service = FactDatasourceService.instance()
        fd_codes = ['test_source_1', 'test_source_2']
        # fd_codes = ['test_source_1', 'test_source_2', 'test_source_x']
        # fd_codes = ['test_source_x']
        is_exists = fd_service.exists_fds(fd_codes)
        print(is_exists)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
from api.model.FactDatasource import FactDatasource, FactDatasourceTypeEnum, JdbcTypeEnum
from factdatasource.FactDatasourceContext import FactDatasourceContext
from factdatasource.dao.FactDatasource import customer_dao_context_holder
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from factdatasource.execptions import NotSupportedFDException, FDNotExistsException, FDOperationException

from libs.utils import get_jdbc_type, Singleton
from service.FactDatasourceService import FactDatasourceService


class FactDatasourceContextHolder(Singleton):

    # fd_code ---> FactDatasourceContext
    __fact_datasource_context_cache = dict()

    # fd_code ---> FactDatasource
    __fact_datasource_cache = dict()

    @property
    def _fd_service(self):
        return FactDatasourceService.instance()

    def get_fact_datasource(self, fd_code: str) -> FactDatasourceContext:
        """
        获取一个fd_context的实现
        这里逻辑应该是这样，先通过rest接口取到fd的定义信息，然后根据这个定义信息构造fd_context
        :param fd_code:
        :return:
        """
        if fd_code in self.__fact_datasource_context_cache.keys():
            return self.__fact_datasource_context_cache.get(fd_code)

        if fd_code not in self.__fact_datasource_cache.keys():
            fact_datasource: FactDatasource = self._fd_service.get_fd_by_code(fd_code)
            if not fact_datasource:
                raise FDNotExistsException(f'数据源{fd_code}不存在')
        else:
            fact_datasource = self.__fact_datasource_cache.get(fd_code)

        if self.add_fact_datasource(fact_datasource):
            return self.__fact_datasource_context_cache.get(fd_code)

    def remove_fact_datasource(self, fact_datasource: FactDatasource):
        """
        删除一个fd实现
        这里只有消息监听处会调用，其他地方暂时没有调用这里的
        :param fact_datasource:
        :return:
        """
        fd_code = fact_datasource.code
        if self.is_using_fact_datasource(fd_code):
            raise FDOperationException(f'数据源{fd_code}正在使用，无法删除。')

        if fd_code in self.__fact_datasource_context_cache.keys():
            del self.__fact_datasource_context_cache[fd_code]
        if fd_code in self.__fact_datasource_cache.keys():
            del self.__fact_datasource_cache[fd_code]

        # 删除数据源中的配置
        if self._need_datasource(fact_datasource.type):
            get_multiple_datesource(fact_datasource.type).delete_datasource_type(fact_datasource.get_source_name())
        return True

    def add_fact_datasource(self, fact_datasource: FactDatasource):
        """
        增加一个fd实现
        :param fact_datasource:
        :return:
        """
        if fact_datasource is None or fact_datasource.code is None:
            return False

        fd_code = fact_datasource.code
        # 放到缓存中
        if fd_code not in self.__fact_datasource_cache.keys():
            self.__fact_datasource_cache[fd_code] = fact_datasource

        fact_datasource_context = self.__get_fact_datasource_context(fact_datasource)
        if fact_datasource_context:
            self.__fact_datasource_context_cache[fd_code] = fact_datasource_context
            # 添加到数据源中
            if self._need_datasource(fact_datasource.type):
                get_multiple_datesource(fact_datasource.type).add_datasource_type(fact_datasource.get_source_name(),
                                                                                  fact_datasource.connection,
                                                                                  fact_datasource.credentials)
            return True
        return False

    def update_fact_datasource(self, fact_datasource: FactDatasource):
        """
        FD修改
        :param fact_datasource:
        :return:
        """
        fd_code = fact_datasource.code
        if self.is_using_fact_datasource(fd_code):
            raise FDOperationException(f'数据源{fd_code}正在使用，无法修改。')
        self.remove_fact_datasource(fact_datasource)
        self.add_fact_datasource(fact_datasource)

    def is_loaded_fact_datasource(self, fd_code: str) -> bool:
        """
        判断fd_code是否已经加载
        :param fd_code:
        :return:
        """
        return fd_code in self.__fact_datasource_cache.keys()

    def is_using_fact_datasource(self, fd_code: str) -> bool:
        """
        判断fd_code是否正在使用，正在使用表示正在读写数据或者查询数据
        :param fd_code:
        :return:
        """
        # 没在内存中肯定就没有被使用
        if not self.is_loaded_fact_datasource(fd_code):
            return False

        fd: FactDatasource = self.__fact_datasource_cache.get(fd_code)
        return customer_dao_context_holder.is_using(fd.get_source_name())

    def _need_datasource(self, fd_type: FactDatasourceTypeEnum):
        """
        判断某一个fd是否需要使用数据源，比如sobeyhvie这种直接调用rest接口的，就不需要使用数据源
        默认都应该使用的
        :param fd_type:
        :return:
        """
        if not fd_type:
            True
        if fd_type == FactDatasourceTypeEnum.SOBEY_HIVE:
            return False
        return True

    def __get_fact_datasource_context(self, fact_datasource: FactDatasource) -> FactDatasourceContext:
        fd_type = fact_datasource.type
        fd_context = None

        if fd_type == FactDatasourceTypeEnum.JDBC:
            # 根据url区分需要选择的方言类型
            connection = fact_datasource.connection
            jdbc_type = get_jdbc_type(connection)
            if not jdbc_type:
                raise NotSupportedFDException(f'不支持该连接{connection}')

            if jdbc_type == JdbcTypeEnum.MYSQL:
                from factdatasource.persistence.jdbc.dialect.MysqlJdbcFactDatasourceContext import \
                    MysqlJdbcFactDatasourceContext
                fd_context = MysqlJdbcFactDatasourceContext(fact_datasource)
            elif jdbc_type == JdbcTypeEnum.DB2:
                from factdatasource.persistence.jdbc.dialect.Db2JdbcFactDatasourceContext import \
                    Db2JdbcFactDatasourceContext
                fd_context = Db2JdbcFactDatasourceContext(fact_datasource)
            elif jdbc_type == JdbcTypeEnum.ORACLE:
                from factdatasource.persistence.jdbc.dialect.OracleJdbcFactDatasourceContext import \
                    OracleJdbcFactDatasourceContext
                fd_context = OracleJdbcFactDatasourceContext(fact_datasource)
            elif jdbc_type == JdbcTypeEnum.POSTGRESQL:
                from factdatasource.persistence.jdbc.dialect.PostgresqlJdbcFactDatasourceContext import \
                    PostgresqlJdbcFactDatasourceContext
                fd_context = PostgresqlJdbcFactDatasourceContext(fact_datasource)
            elif jdbc_type == JdbcTypeEnum.SQL_SERVER:
                from factdatasource.persistence.jdbc.dialect.SqlserverJdbcFactDatasourceContext import \
                    SqlserverJdbcFactDatasourceContext
                fd_context = SqlserverJdbcFactDatasourceContext(fact_datasource)
            elif jdbc_type == JdbcTypeEnum.VERNOX:
                from factdatasource.persistence.jdbc.dialect.VernoxJdbcFactDatasourceContext import \
                    VernoxJdbcFactDatasourceContext
                fd_context = VernoxJdbcFactDatasourceContext(fact_datasource)
            else:
                raise NotSupportedFDException(f'暂不支持该数据库{jdbc_type}')
        elif fd_type == FactDatasourceTypeEnum.AMQP:
            from factdatasource.persistence.amqp.AmqpFactDatasourceContext import AmqpFactDatasourceContext
            fd_context = AmqpFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.CUSTOM:
            from factdatasource.persistence.custom.CustomFactDatasourceContext import CustomFactDatasourceContext
            fd_context = CustomFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.ES:
            from factdatasource.persistence.es.EsFactDatasourceContext import EsFactDatasourceContext
            fd_context = EsFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.FILE:
            from factdatasource.persistence.file.FileFactDatasourceContext import FileFactDatasourceContext
            fd_context = FileFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.GRAPH:
            from factdatasource.persistence.graphdb.GraphdbFactDatasourceContext import GraphdbFactDatasourceContext
            fd_context = GraphdbFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.JMS:
            from factdatasource.persistence.stomp.StompFactDatasourceContext import StompFactDatasourceContext
            fd_context = StompFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.KAFKA:
            from factdatasource.persistence.kafka.KafkaFactDatasourceContext import KafkaFactDatasourceContext
            fd_context = KafkaFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.MONGO:
            from factdatasource.persistence.mongo.MongoFactDatasourceContext import MongoFactDatasourceContext
            fd_context = MongoFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.REDIS:
            from factdatasource.persistence.redis.RedisFactDatasourceContext import RedisFactDatasourceContext
            fd_context = RedisFactDatasourceContext(fact_datasource)
        elif fd_type == FactDatasourceTypeEnum.SOBEY_HIVE:
            from factdatasource.persistence.sobeyhive.SobeyHiveFactDatasourceContext import SobeyHiveFactDatasourceContext
            fd_context = SobeyHiveFactDatasourceContext(fact_datasource)
        else:
            raise NotSupportedFDException(f'暂不支持该FD类型{fd_type}')

        return fd_context




server_ip = None
server_port = 5000
config_name = "sobey-cube"
config_profile = "release"
config_server_id = "sobeyficus-config-server"
config_fail_fast = True
eureka_default_zone = "http://sobeyficus:JXTYp9icQaTzs4@sobeyficus-eureka:8765/eureka/"
application_name = None
actor_name = None
spring_profiles_active = "default"

def find_host_ip():
    """
    获取本机的IP地址
    :return:
    """
    global s
    try:
        # 创建一个临时的socket连接
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 向自己发送一个udp的请求
        s.connect(('8.8.8.8', 80))
        # 送结果中获取到真实的ip地址
        ip = s.getsockname()[0]
    finally:
        s.close()

    import config
    if config.server_ip is None:
        config.server_ip = ip
    return ip


"""
CE的操作模块
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def get(site, project_code, code):
    """
    获取CE
    :param site: 站点
    :param project_code: 项目名
    :param code: CE的唯一code
    :return: 如果这个ce存在,返回ce对象.否则返回空
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/ce-service/{site}/{project_code}/{code}")

        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def exists(site, project_code, code):
    """
    校验CE是否存在
    :param site: 站点
    :param project_code: 项目名
    :param code: CE的唯一code
    :return: 如果这个ce存在,返回true.否则,返回false
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/ce-service/exists/{site}/{project_code}/{code}")

        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def add_compute_execution(data_compute_execution):
    """
    新增一个CE
    :param data_compute_execution: ce对象
    :return: 如果添加失败,就抛出异常,否则无返回
    """
    if data_compute_execution is None:
        return

    check_instance_avaliable()

    request = None
    if not isinstance(data_compute_execution, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(data_compute_execution, dict):
            # 如果是dict类型的,直接发
            request = data_compute_execution
    else:
        # 说明是munch的,那么就转成Dict的
        request = data_compute_execution.toDict()

    try:
        r = do_service(f"remote/ce-service/",method="post",data=request,return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



from abc import abstractmethod

from api.model.ResultVO import ResultVO
from api.model.AsyncServiceResponse import AsyncServiceResponse

class IAsyncAble:

    @abstractmethod
    def do_finish(self, task_log_id: int, response, header: dict, ficus_param: dict, task_status: int) -> ResultVO:
        pass

    @abstractmethod
    def transform_response(self, response, headers: dict) -> AsyncServiceResponse:
        pass


from enum import Enum


class SendType(Enum):
    NEED_ASYNC_SEND = 'NEED_ASYNC_SEND'  # 需要发送异步任务给执行器
    NEED_SYNC_SEND = 'NEED_SYNC_SEND'  # 需要发送同步任务给执行器
    NOT_NEED_SEND = 'NOT_NEED_SEND'  # 不需要发送任务给执行器


class AcceptType(Enum):
    APPLICATION_JSON = 'APPLICATION_JSON'
    TEXT_PLAIN = 'TEXT_PLAIN'


class AsyncServiceRequest:
    """
    异步的请求
    """

    def __init__(self, url: str, method: str = 'POST', header: dict = None, body=None,
                 send_type: SendType = SendType.NEED_ASYNC_SEND, accept_type: AcceptType = AcceptType.APPLICATION_JSON):
        """

        :param url: 请求的URL地址
        :param method:  请求方法
        :param header:  消息头
        :param body:    消息体
        :param send_type:   默认为需要发送异步任务给执行器
        :param accept_type:  接收回调的类型
        """
        self.url = url
        self.method = method
        self.header = header
        self.body = body
        self.send_type = send_type
        self.accept_type = accept_type



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/23
import json

from pika import BlockingConnection

from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton, JSONEncoder


class AmqpDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.AMQP)
        return data_source

    @property
    def client(self) -> BlockingConnection:
        return self.datasource.get_client()

    def basic_publish_topic(self, exchange: str, routing_key: str, message):
        """
        发送topic消息
        :param exchange:
        :param routing_key:
        :param message:可以一次发送多个消息，多个消息共享一个channel
        :type message: str or list
        :return:
        """
        with self.client.channel() as channel:
            channel.exchange_declare(exchange=exchange, exchange_type='topic')
            if isinstance(message, list):
                for mess in message:
                    if not isinstance(mess, str):
                        mess = json.dumps(mess, cls=JSONEncoder)
                        channel.basic_publish(exchange=exchange, routing_key=routing_key, body=mess)
                    else:
                        channel.basic_publish(exchange=exchange, routing_key=routing_key, body=mess)
            else:
                if not isinstance(message, str):
                    message = json.dumps(message, cls=JSONEncoder)
                channel.basic_publish(exchange=exchange, routing_key=routing_key, body=message)



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="FicusFramework",
    version="3.0.6.post3",

    packages=find_packages('src'),
    package_dir={'': 'src'},
    test_suite="test",

    install_requires=["flask==1.1.2", "flask-cors==3.0.8", "requests==2.23.0", "apscheduler==3.6.3", "jsonpath-rw==1.4.0",
                      "munch==2.5.0", "PyYaml==5.3.1", "readerwriterlock==1.0.7", "confluent-kafka==1.3.0",
                      "sqlalchemy==1.3.3", "mysqlclient","gevent","Celery==4.4.2","redis==3.5.3"],

    author='sunxiang0918',
    author_email="sunxiang0918@gmail.com",
    description="A framework for Ficus by Python3.",
    long_description=read("README.rst"),
    license="MIT",
    keywords="ficus framework python",
    url="https://git.sobey.com/SobeyHive/FicusFramework4Py",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Natural Language :: Chinese (Simplified)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ],
    zip_safe=False
)



from api.exceptions import UnSupportException
from rule_engine import IRuleAction


class DefaultAction(IRuleAction):
    """
    默认的action就是把 action中的东西返回出去
    """

    def do_action(self, action, facts, params: dict):
        if params is None:
            params = {}

        params["_action"] = action

        return params

class UnSupportAction(IRuleAction):
    def do_action(self, action, facts, params: dict):
        raise UnSupportException("此执行动作Python 并不支持")

class PyExpAction(IRuleAction):
    """
    python表达式的方式,类似于JAVA中的MVEL
    """
    def do_action(self, action, facts, params: dict):
        from munch import Munch

        context = facts.get_facts() or {}
        if params is not None:
            context.update(params)
        context = Munch.fromDict(context)

        params_ = Munch.fromDict(context)

        exec(f"def __b(params):\n{self.__reindent(action,1)}\n", context)
        result2 = context["__b"]
        return result2(params_)

    def __reindent(self,s, numSpaces):
        """
        多行代码的统一缩进处理
        :param s:
        :param numSpaces:
        :return:
        """
        leading_space = numSpaces * ' '
        lines = [leading_space + line for line in s.splitlines()]
        return '\n'.join(lines)


from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer


class CompleteGB32100Transformer(ITransformer):
    """
    社会统一代码补全 先不做
    """

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):
        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # TODO 这个先不实现

        return {target_model_field.fieldName: data[target_model_field.fieldName]}



from abc import abstractmethod

from client import ScheduleJobTaskLogClient


class IProgressAble:

    @abstractmethod
    def get_task_log_id(self):
        """
        获取任务的ID
        :return:
        """
        pass

    def update_task_progress(self, progress: float):
        """
        更新任务的进度
        :return:
        """
        task_log_id = self.get_task_log_id()
        if task_log_id is None:
            return

        ScheduleJobTaskLogClient.update_task_progress(task_log_id, progress)



import requests

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def count_tasks(task_type, status=[]) -> int:
    """
    获取某一个任务在某些状态下的数量
    :param task_type:
    :param status:
    :return:
    """
    if task_type is None or status is None:
        return 0

    check_instance_avaliable("SOBEYCLOUD-SCHEDULE-SERVICE")

    try:
        r = do_service(
            f"/v1.0/S1/tasks/count?task_type={task_type}" + (
                ("&" + "&".join(map(lambda x: f"status={x}", status))) if len(status) > 0 else ""),
            app_name="SOBEYCLOUD-SCHEDULE-SERVICE", method="get", return_type="int")
        return r
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



from abc import abstractmethod

from api.handler import HandlerEnum
from api.handler.IHandler import IHandler
from api.handler.script.ICacheAbleScript import ICacheAbleScript
from api.handler.IKillable import IKillable
from api.handler.ILogAbleHandler import ILogAbleHandler
from api.handler.script.IProgressAbleScript import IProgressAbleScript


class ISimpleScriptCrawl(IKillable, ILogAbleHandler, ICacheAbleScript, IHandler, IProgressAbleScript):
    """
    Python脚本式的Crawl的基类
    """

    @abstractmethod
    def do_crawl(self, params: dict):
        """
        抓取的业务逻辑
        :return: 返回一个 OutputWrapper的 数组
        """
        pass

    def handler_enum(self) -> HandlerEnum:
        return HandlerEnum.SCRIPT

    def is_killed(self) -> bool:
        """
        判断是否杀掉了
        :return:
        """
        return False



import logging

from api.exceptions import IllegalArgumentException, ServiceInnerException
from rule_engine import RuleFacts, RuleGroupInfo, RuleResult
from rule_engine.core import IRuleGroupSession
from rule_engine.core.DefaultAgenda import DefaultAgenda

log = logging.getLogger('Ficus')


class SimpleRuleGroupSession(IRuleGroupSession):

    def __init__(self, rule_group_info:RuleGroupInfo) -> None:
        self.__added_fact = False
        self.__rule_listeners = []
        self.__rule_facts = None
        self.__agenda = DefaultAgenda(rule_group_info,self.__rule_listeners)

    def add_facts(self, rule_facts: RuleFacts):
        if rule_facts is None:
            self.__rule_facts = RuleFacts()

        if self.__rule_facts is None:
            self.__rule_facts = rule_facts
        else:
            self.__rule_facts.add_facts(rule_facts)

        self.__added_fact = True

    def add_fact(self, key: str, value):
        if self.__rule_facts is None:
            self.__rule_facts = RuleFacts()

        self.__rule_facts.add_fact(key,value)

        self.__added_fact = True


    def add_rule_listener(self, rule_listener):
        self.__rule_listeners.append(rule_listener)

    def execute(self):
        if not self.__added_fact:
            raise IllegalArgumentException("未添加规则事实,无法进行规则匹配")

        log.info(f"开始一次规则的匹配,{self.__agenda}")

        log.debug(f"规则事实:{self.__rule_facts}")

        # 把事实传入.然后匹配出符合条件的规则.
        infos = self.__agenda.get_matched_rules(self.__rule_facts)

        if infos is None or len(infos)==0:
            log.info(f"没有匹配到规则,{self.__agenda}")
            return None

        results = []

        log.info(f"成功匹配到规则,{self.__agenda}")

        log.debug(f"匹配到的规则:{infos}")

        for info in infos:
            # 执行结果
            try:
                if not self.__triggerListenersBeforeExecute(info,self.__rule_facts):
                    log.info(f"规则 {info.name} 跳过执行action阶段")
                    continue

                rule_result = RuleResult(info.name)
                results.append(rule_result)

                then = info.then
                if then is not None and len(then)>0:
                    for rule_action in then:
                        action = rule_action.rule_action.do_action(rule_action.action,self.__rule_facts,rule_action.params)
                        rule_result.add_result(action)
                self.__triggerListenersOnSuccess(info,self.__rule_facts,rule_result)
            except Exception as e:
                log.error(f"执行匹配的规则 {info.name} 失败,{str(e)}")
                self.__triggerListenersOnFailure(info,e,self.__rule_facts)
                raise ServiceInnerException(f"执行匹配的规则 {info.name} 失败,{str(e)}")

        return results

    # region 消息相关
    def __triggerListenersBeforeExecute(self, rule, rule_facts):
        for rule_listener in  self.__rule_listeners:
            if not rule_listener.before_action(rule,rule_facts):
                return False

        return True

    def __triggerListenersOnSuccess(self, rule, rule_facts, rule_result):
        for rule_listener in self.__rule_listeners:
            rule_listener.on_success(rule,rule_facts,rule_result)

    def __triggerListenersOnFailure(self, rule, e, rule_facts):
        for rule_listener in self.__rule_listeners:
            rule_listener.on_failure(rule, rule_facts, e)
    # endregion


import logging

from celery.result import AsyncResult

from api.model.ResultVO import ResultVO, FAIL_CODE
from cloudcelery import celery
from schedule import TriggerActor
from munch import Munch
import json

log = logging.getLogger('Ficus')


@celery.task(name='tasks.on_request', bind=True, max_retries=2, default_retry_delay=1 * 6)
def on_request(self, protocol):
    """
    从celery接收协议
    :param self:
    :param protocol:
    :return:
    """
    log.info(f"从celery中获取到任务:{protocol}")
    try:
        # 子进程中也开启eureka_client的刷新,因为celery也会起子进程,所以只能在这里初始化了
        from py_eureka_client import eureka_client
        import config
        eureka_client.init_discovery_client(eureka_server=config.eureka_default_zone, renewal_interval_in_secs=4)
    except:
        pass

    body = Munch(json.loads(protocol))

    result: ResultVO = TriggerActor.handle_trigger(body, True)  # ResultVO

    if result.code == FAIL_CODE:
        # 让celery的任务也置失败
        raise RuntimeError(result.msg)

    return {'status': True, 'data': result.to_dict()}



from datetime import datetime

from api.model.MetaModelField import MetaModelField, DataTypeEnum
from factdatasource.transformer.ITransformer import ITransformer

DATE_FORMAT_KEY = "__dateFormat__"
DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
DEFAULT_VALUE_KEY = "__default_value__"
NOW = "now()"


class DefaultValueTransformer(ITransformer):
    """
    默认值的转换器,
    """

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):
        if (params is None or len(params) == 0) or (
                DEFAULT_VALUE_KEY not in params or params[DEFAULT_VALUE_KEY] is None or params[
            DEFAULT_VALUE_KEY] == ""):

            # 没有设置默认值,不管他
            if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
                return {target_model_field.fieldName: None}

            return data[target_model_field.fieldName]

        old_value = data[target_model_field.fieldName]

        # 默认值,String的
        default_string = params[DEFAULT_VALUE_KEY]

        # 需要根据字段类型进行变更
        # 找到目标字段类型
        data_type = target_model_field.dataType

        if data_type == DataTypeEnum.STRING:
            return {target_model_field.fieldName: default_string if old_value is None else old_value}
        elif data_type in [DataTypeEnum.FLOAT, DataTypeEnum.LONG, DataTypeEnum.INTEGER, DataTypeEnum.DOUBLE]:
            return {target_model_field.fieldName: self.__inner_number_convert(old_value, data_type, default_string)}
        elif data_type == DataTypeEnum.DATE:
            return {target_model_field.fieldName: self.__inner_date_convert(old_value,
                                             DEFAULT_DATE_FORMAT if DATE_FORMAT_KEY not in params else params[
                                                 DATE_FORMAT_KEY], default_string)}
        elif data_type == DataTypeEnum.BOOLEAN:
            return {target_model_field.fieldName: self.__inner_boolean_convert(old_value, default_string)}
        elif data_type == DataTypeEnum.CHAR:
            return {target_model_field.fieldName: self.__inner_char_convert(old_value, default_string)}
        elif data_type == DataTypeEnum.BYTE:
            return {target_model_field.fieldName: self.__inner_byte_convert(old_value, default_string)}

    def __inner_number_convert(self, old_value, data_type, default_string):
        if old_value is not None:
            return old_value

        if data_type == DataTypeEnum.Long or data_type == DataTypeEnum.INTEGER:
            return int(default_string)
        elif data_type == DataTypeEnum.FLOAT or data_type == DataTypeEnum.DOUBLE:
            return float(default_string)

        return None

    def __inner_date_convert(self, old_value, date_format, default_string):
        if old_value is not None:
            return old_value

        if NOW == default_string:
            return datetime.now().strftime(DEFAULT_DATE_FORMAT)
        else:
            return datetime.strptime(default_string, date_format)

    def __inner_boolean_convert(self, old_value, default_string):
        if old_value is not None:
            return old_value

        return bool(default_string)

    def __inner_char_convert(self, old_value, default_string):
        if old_value is not None:
            return old_value

        return default_string[0]

    def __inner_byte_convert(self, old_value, default_string: str):
        if old_value is not None:
            return old_value

        if default_string.lower().startswith("0x"):
            return int(default_string[2:])
        else:
            return int(default_string)



import logging
import config

log = logging.getLogger('Ficus')


def load_properties_after_started(app):
    """
    加载配置文件  bootstrap.yml
    :param app:
    :return:
    """
    # 1. 获取yaml文件
    if _exists_bootstrap_property(app) is None:
        return

    # 这里开始尝试访问cloud-config获取配置文件
    from client import check_instance_avaliable
    from api.exceptions import ServiceNoInstanceException
    try:
        check_instance_avaliable(config.config_server_id)
    except ServiceNoInstanceException:
        if config.config_fail_fast:
            from api.exceptions import ServiceNoInstanceException
            raise ServiceNoInstanceException("cloud-config服务没有启动,系统启动失败")
        else:
            log.info("没有从注册中心上找到配置中心,使用默认配置")
        return

    # region 发起请求 获取配置,并设置到 annotation.REMOTE_YML_CONFIG 中
    from urllib.error import HTTPError
    try:
        # 支持配置多个
        configs = config.config_name.split(",")
        from client import do_service
        for conf in configs:
            r = do_service(f"{conf}-{config.config_profile}.yml", app_name=config.config_server_id,
                           return_type="str", auth=False)  # 这里是从配置中心去获取数据,不做auth
            # 配置中心的配置文件不存在时，返回的是 '{}'
            if r is not None and str(r).rstrip() != '{}':
                from config import annotation
                import yaml
                # 合并本地配置文件和云端配置文件
                annotation.deep_search_and_merge(annotation.REMOTE_YML_CONFIG,
                                                 yaml.load(r, Loader=yaml.FullLoader))  # r.content
                log.info(f"完成配置中心配置读取{conf}-{config.config_profile}")
            else:
                from api.exceptions import ServiceInnerException
                raise ServiceInnerException(f"cloud-config服务没有 {conf}-{config.config_profile}.yml 配置")
    except HTTPError:
        if config.config_fail_fast:
            from api.exceptions import ServiceNoInstanceException
            raise ServiceNoInstanceException("cloud-config服务没有启动,系统启动失败")
        else:
            # 如果不快速失败,就直接返回
            return
    # endregion


def _exists_bootstrap_property(app):
    from registry import read_yaml_file
    yml = read_yaml_file(app, "bootstrap.yml", config.spring_profiles_active)
    return yml


def init_from_yaml_property(app):
    """
    读取本地的配置文件
    :param app:
    :return:
    """

    # 先读取 applicaton.yml 文件
    _read_from_application_yaml_property(app)

    # 再读取bootstrap.yml 文件
    yml = _exists_bootstrap_property(app)

    if yml is None:
        # 说明没有yml文件,不进行注册处理
        log.info("没有配置bootstrap.yml,采用默认的配置信息")
        return False

    # 说明有yml文件,开始进行处理

    # 第一个: server.port
    if "server" in yml and "port" in yml["server"]:
        config.server_port = yml["server"]["port"]

    if "server" in yml and "ip" in yml["server"]:
        config.server_ip = yml["server"]["ip"]

    try:
        config.config_name = yml["spring"]["cloud"]["config"]["name"]
    except:
        pass

    try:
        config.config_profile = yml["spring"]["cloud"]["config"]["profile"]
    except:
        pass

    try:
        config.config_server_id = yml["spring"]["cloud"]["config"]["discovery"]["service-id"]
    except:
        pass

    try:
        config.config_fail_fast = yml["spring"]["cloud"]["config"]["fail-fast"]
    except:
        pass

    try:
        config.application_name = yml["spring"]["application"]["name"]
    except:
        pass

    try:
        config.eureka_default_zone = yml["eureka"]["client"]["service-url"]["defaultZone"]
    except:
        pass

    return True


def init_profile_from_environ_property():
    import os

    environs = os.environ

    if "spring.profiles.active" in environs:
        config.spring_profiles_active = environs["spring.profiles.active"]
        log.info(f"从系统环境变量中获取到spring.profiles.active:{config.spring_profiles_active}")
    elif "SPRING_PROFILES_ACTIVE" in environs:
        config.spring_profiles_active = environs["SPRING_PROFILES_ACTIVE"]
        log.info(f"从系统环境变量中获取到SPRING_PROFILES_ACTIVE:{config.spring_profiles_active}")


def init_from_environ_property():
    import os

    environs = os.environ
    # 第一个: server.port
    if "server.port" in environs:
        config.server_port = int(environs["server.port"])
        log.info(f"从系统环境变量中获取到server.port:{config.server_port}")
    elif "SERVER_PORT" in environs:
        config.server_port = int(environs["SERVER_PORT"])
        log.info(f"从系统环境变量中获取到SERVER_PORT:{config.server_port}")

    if "server.ip" in environs:
        config.server_ip = environs["server.ip"]
        log.info(f"从系统环境变量中获取到server.ip:{config.server_ip}")
    elif "SERVER_IP" in environs:
        config.server_ip = environs["SERVER_IP"]
        log.info(f"从系统环境变量中获取到SERVER_IP:{config.server_ip}")

    if "spring.cloud.config.name" in environs:
        config.config_name = environs["spring.cloud.config.name"]
        log.info(f"从系统环境变量中获取到spring.cloud.config.name:{config.config_name}")
    elif "SPRING_CLOUD_CONFIG_NAME" in environs:
        config.config_name = environs["SPRING_CLOUD_CONFIG_NAME"]
        log.info(f"从系统环境变量中获取到SPRING_CLOUD_CONFIG_NAME:{config.config_name}")

    if "spring.cloud.config.profile" in environs:
        config.config_profile = environs["spring.cloud.config.profile"]
        log.info(f"从系统环境变量中获取到spring.cloud.config.profile:{config.config_profile}")
    elif "SPRING_CLOUD_CONFIG_PROFILE" in environs:
        config.config_profile = environs["SPRING_CLOUD_CONFIG_PROFILE"]
        log.info(f"从系统环境变量中获取到SPRING_CLOUD_CONFIG_PROFILE:{config.config_profile}")

    if "spring.cloud.config.fail-fast" in environs:
        config.config_fail_fast = bool(environs["spring.cloud.config.fail-fast"])
        log.info(f"从系统环境变量中获取到spring.cloud.config.fail-fast:{config.config_fail_fast}")
    elif "SPRING_CLOUD_CONFIG_FAILFAST" in environs:
        config.config_fail_fast = bool(environs["SPRING_CLOUD_CONFIG_FAILFAST"])
        log.info(f"从系统环境变量中获取到SPRING_CLOUD_CONFIG_FAILFAST:{config.config_fail_fast}")

    if "spring.cloud.config.discovery.service-id" in environs:
        config.config_server_id = environs["spring.cloud.config.discovery.service-id"]
        log.info(f"从系统环境变量中获取到spring.cloud.config.discovery.service-id:{config.config_server_id}")
    elif "SPRING_CLOUD_CONFIG_DISCOVERY_SERVICEID" in environs:
        config.config_server_id = environs["SPRING_CLOUD_CONFIG_DISCOVERY_SERVICEID"]
        log.info(f"从系统环境变量中获取到SPRING_CLOUD_CONFIG_DISCOVERY_SERVICEID:{config.config_server_id}")

    if "spring.application.name" in environs:
        config.application_name = environs["spring.application.name"]
        log.info(f"从系统环境变量中获取到spring.application.name:{config.application_name}")
    elif "SPRING_APPLICATION_NAME" in environs:
        config.application_name = environs["SPRING_APPLICATION_NAME"]
        log.info(f"从系统环境变量中获取到SPRING_APPLICATION_NAME:{config.application_name}")

    if "eureka.client.service-url.defaultZone" in environs:
        config.eureka_default_zone = environs["eureka.client.service-url.defaultZone"]
        log.info(f"从系统环境变量中获取到eureka.client.service-url.defaultZone:{config.eureka_default_zone}")
    elif "EUREKA_CLIENT_SERVICEURL_DEFAULTZONE" in environs:
        config.eureka_default_zone = environs["EUREKA_CLIENT_SERVICEURL_DEFAULTZONE"]
        log.info(f"从系统环境变量中获取到EUREKA_CLIENT_SERVICEURL_DEFAULTZONE:{config.eureka_default_zone}")


def _read_from_application_yaml_property(app):
    """
    读取本地的文件到系统配置中
    :param app:
    :return:
    """
    from registry import read_yaml_file
    yml = read_yaml_file(app, "application.yml", config.spring_profiles_active)
    if yml is None:
        # 说明没有yml文件,不进行注册处理
        log.info("没有配置application.yml,采用默认的配置信息")
        return
    from config import annotation
    # 把本地的信息加入到配置中
    annotation.REMOTE_YML_CONFIG = yml

    try:
        config.actor_name = yml["actor"]["name"]
    except:
        pass



from enum import Enum, unique


@unique
class OperationEnum(Enum):
    INSERT = 0
    UPDATE = 1
    UPSERT = 2
    CLEAN = 3
    DELETE = 4
    FLUSH = 5


class OutputWrapper:
    """
    数据操作包装类
    """
    __index = 0

    __content = None

    __operation: OperationEnum = None

    def __init__(self, content, operation, index):
        self.__content = content
        self.__operation = operation
        self.__index = index

    def content(self):
        return self.__content

    def operation(self):
        return self.__operation

    def index(self):
        return self.__index


def INSERT(content, index=0):
    """
    插入操作一个内容
    :param content: 被操作的内容
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(content, OperationEnum.INSERT, index)


def UPDATE(content, index=0):
    """
    更新操作一个内容
    :param content: 被操作的内容
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(content, OperationEnum.UPDATE, index)


def UPSERT(content, index=0):
    """
    saveOrUpdate操作一个内容
    :param content: 被操作的内容
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(content, OperationEnum.UPSERT, index)


def DELETE(query, index=0):
    """
    删除操作一个内容
    :param query: 被删除的查询条件
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(query, OperationEnum.DELETE, index)


def CLEAN(index=0):
    """
    清除操作
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(None, OperationEnum.CLEAN, index)

def FLUSH(index=0):
    """
    强制提交缓存
    :param index: 操作层级,一般不管
    :return: OutputWrapper对象
    """
    return OutputWrapper(None, OperationEnum.FLUSH, index)


from abc import abstractmethod

from api.model.MetaModelField import MetaModelField


class ITransformer:
    """
    数据字段的转换接口
    """

    @abstractmethod
    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField) -> dict:
        """
        字段的值转换
        :param data: 源数据,
        :param params: 可能的转换参数
        :param target_model_field: 模型字段
        :param default_value: 默认值
        :return: 转换的结果, key是字段的名字, value是转换后的结果, 这里是一个Map的结果主要原因是,某一些转换器会把一个字段的结果转换成为多个字段,相当于是字段的分裂.因此需要使用Map来保存
        """
        pass



"""
缓存相关的操作
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def set_value(key, value):
    """
    设置缓存值
    :param key:
    :param value:
    :return:
    """
    if key is None or value is None:
        return

    check_instance_avaliable()

    request = value
    if not isinstance(value, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(value, dict):
            # 如果是dict类型的,直接发
            request = value
    else:
        # 说明是munch的,那么就转成Dict的
        request = value.toDict()

    try:
        r= do_service(f"remote/sc-service/{key}", data=request, method="post",return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def set_if_absent(key, value):
    """
    设置缓存自,如果key不存在
    :param key:
    :param value:
    :return:
    """
    if key is None or value is None:
        return False

    check_instance_avaliable()

    request = None
    if not isinstance(value, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(value, dict):
            # 如果是dict类型的,直接发
            request = value
    else:
        # 说明是munch的,那么就转成Dict的
        request = value.toDict()

    try:
        r = do_service(f"remote/sc-service/{key}", data=request, method="put")
        return r
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def get(key):
    """
    获取某一个缓存
    :param key:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sc-service/{key}")
        if r is not None:
            try:
                return Munch(r)
            except (TypeError,ValueError,AttributeError):
                return r
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def delete(key):
    """
    删除某一个缓存
    :param key:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/sc-service/{key}",method="delete",return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/6
import logging
import sys
import time
from abc import abstractmethod
from unittest import TestCase

from munch import Munch


class BaseInitTestSuite(TestCase):
    client = None

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    def distributed_mode(self) -> bool:
        """
        是否使用分布式的FD，默认使用， 如果要使用ficus中心模式的FD，覆盖此方法
        :return:
        """
        return True

    def centralized_mode_config_path(self) -> str:
        """
        ficus中心模式时，配置文件夹路径,如果sys.path[0]不能满足路径配置，覆盖此方法
        :return:
        """
        return sys.path[0]

    def execute(self, target, param):
        print(f'开始对{param.get("code_")}进行测试')
        start = time.time()
        try:
            from api.model.ResultVO import ResultVO
            result: ResultVO = target.execute(param)
            print(f'测试结果:{result.msg}')
            # 验证结果
            from api.model.ResultVO import SUCCESS
            self.assertEqual(result.code, SUCCESS.code)
        finally:
            print(f'测试结束用时{time.time()-start}')

    def _get_datesource_client(self):
        if self.client is not None:
            return self.client

        from config.BootstrapPropertyLoader import init_from_yaml_property,init_from_environ_property
        config_path = self.centralized_mode_config_path()
        init_from_yaml_property(config_path)
        # 尝试从环境变量中获取 bootstrap里面的信息
        init_from_environ_property()

        if self.distributed_mode():
            from factdatasource.FactDatasourceProxyService import DistributedFactDatasourceProxy
            self.client = DistributedFactDatasourceProxy.instance()
            # 注册所有的数据源
            self.__register_fact_datasource()
        else:
            # from config.BootstrapPropertyLoader import init_from_yaml_property
            # init_from_yaml_property(self.centralized_mode_config_path())
            from init import app
            app.log.info('初始化加载配置')

            from factdatasource.FactDatasourceProxyService import CentralizedFactDatasourceProxy
            self.client = CentralizedFactDatasourceProxy.instance()
        return self.client

    def __register_fact_datasource(self):
        fds = self.mock_fact_datasource()
        from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
        fd_context_holder = FactDatasourceContextHolder.instance()
        for fd in fds.values():
            fd_context_holder.add_fact_datasource(fd)


class ScheduleCacheClientMock(object):
    _cache = dict()

    def set_value(self, key, value):
        if key is None or value is None:
            return
        self._cache[key] = value

    def set_if_absent(self, key, value):
        if key is None or value is None:
            return False

        if key in self._cache.keys():
            return False

        self._cache[key] = value

    def get(self, key):
        if key is None:
            return None
        value = self._cache.get(key)
        if value is not None:
            if isinstance(value, dict):
                return Munch(value)
            else:
                return value
        else:
            return None

    def delete(self, key):
        if key is None:
            return
        del self._cache[key]


class DataCrawl(object):
    def __init__(self, site, projectCode, code, outputFdCodes: list, params: dict = None, **kwargs):
        """
        :param site: 站点，一般定义为S1
        :param projectCode:
        :param code: crawl的唯一标示
        :param outputFdCodes:  输出主数据 list类型 list<fd_code>
        :param params: 执行参数
        :param type:  固定值 CUSTOM
        :param kwargs:  其他参数
        """
        self.site = site
        self.projectCode = projectCode
        self.code = code
        self.params = params
        self.outputFdCodes = outputFdCodes
        self.type = 'CUSTOM'
        self.__fields = {
            'site': site,
            'projectCode': projectCode,
            'code': code,
            'outputFdCodes': outputFdCodes,
            'params': params,
            'type': 'CUSTOM'
        }
        self.__fields.update(**kwargs)

    def to_dict(self):
        return self.__fields


class DataComputeExecution(object):
    def __init__(self, site, projectCode, code, sourceFdCodes: list, outputFdCodes: list, params: dict = None, **kwargs):
        """
        :param site: 站点，一般定义为S1
        :param projectCode:
        :param code: crawl的唯一标示
        :param sourceFdCodes:  输入数据源 list类型 list<fd_code>
        :param outputFdCodes:  输出主数据 list类型 list<fd_code>
        :param params: 执行参数
        :param type:  固定值 CUSTOM
        :param kwargs:  其他参数
        """
        self.site = site
        self.projectCode = projectCode
        self.code = code
        self.params = params
        self.sourceFdCodes = sourceFdCodes
        self.outputFdCodes = outputFdCodes
        self.type = 'CUSTOM'
        self.__fields = {
            'site': site,
            'projectCode': projectCode,
            'code': code,
            'sourceFdCodes': sourceFdCodes,
            'outputFdCodes': outputFdCodes,
            'params': params,
            'type': 'CUSTOM'
        }
        self.__fields.update(**kwargs)

    def to_dict(self):
        return self.__fields


class TaskLoggerMock(object):

    def log(self, append_log: str):
        print(append_log)

    def error(self, e: Exception):
        # sys.stderr.write(str(e))
        logging.exception(e)


PROCESS_LOG_ID = "TEST_PROCESS_LOG_ID"
LOG_ID = 0



class ServiceNoInstanceException(Exception):
    pass

class AuthException(Exception):
    pass

class ServiceInnerException(Exception):
    pass

class IllegalArgumentException(Exception):
    pass

class UnSupportException(Exception):
    pass


import threading
from datetime import datetime

from api.handler import HandlerEnum
from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.outputer.ISimpleOutputer import ISimpleOutputer
from api.handler.script import ScriptPythonFactory, ScriptHandlerHolder
from api.model.FdInputPipe import FdInputPipe
from api.model.ResultVO import *
from client import ComputeExecutionClient, DataCrawlClient
from schedule.utils.log import FrameworkHandlerLogger


class ScriptPythonTaskHandler(ITaskHandler,ISimpleOutputer, ICacheAbleHandler):
    """
    Python脚本的执行器
    """

    # 由于这个的实现有可能是 单例的,所以要使用ThreadLocal

    def __init__(self, job_id, update_time, script_source):
        super().__init__()
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__execution_message_local = threading.local()

        self.job_id = job_id
        self.update_time = update_time
        self.script_source = script_source
        self.killed = False
        self.current_task = None

    def update_time(self):
        return self.update_time

    def execute(self, params):
        """
        执行Python脚本的处理
        :param params:
        :return:
        """
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在ce的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有找到执行器的信息")

        self.killed = False

        # 获取ce定义
        dataComputeExecution = ComputeExecutionClient.get(params["site_"], params["projectCode_"], params["code_"])

        if dataComputeExecution is None:
            # 说明有可能是crawl
            dataCrawl = DataCrawlClient.get(params["site_"], params["projectCode_"], params["code_"])
            if dataCrawl is None:
                ResultVO(FAIL_CODE, "执行失败,没有找到执行器的信息")
            return self.do_crawl(dataCrawl, params)

        return self.do_script_compute_execution(dataComputeExecution,params)


    def kill(self):
        self.killed = True

        if  self.current_task is not None:
            try:
                self.current_task.kill()
            except Exception as e:
                FrameworkHandlerLogger.log_error(
                    f"当前正在执行的任务强制停止失败: {str(e)}")

        ScriptPythonFactory.destroy_instance(f"ScriptPython{self.job_id}")

    def is_killed(self):
        return self.killed

    def get_execution_message_cache(self):
        """
        返回message_cache
        :return:
        """
        return self.__execution_message_local.content

    def get_code_thread_local(self):
        return self.__code_local_host

    def get_process_thread_local(self):
        """
        上下文的Id
        :return:
        """
        return self.__process_id_local

    def do_crawl(self, dataCrawl, params):
        """
        执行动态的crawl
        :param dataCrawl:
        :param params:
        :return:
        """
        # 这里动态的加载script源文件
        instance = ScriptPythonFactory.load_instance(f"ScriptPython{self.job_id}", self.script_source,"IHandler")

        if instance.handler_enum() == HandlerEnum.NORMAL:
            return self.__inner_do_normal_crawl(instance,params)
        else:
            return self.__inner_do_script_crawl(instance, dataCrawl, params)

    def do_script_compute_execution(self, dataComputeExecution, params):
        """
        执行ce的脚本
        :param dataComputeExecution:
        :param params:
        :return:
        """

        # 这里动态的加载script源文件
        instance = ScriptPythonFactory.load_instance(f"ScriptPython{self.job_id}", self.script_source,"IHandler")

        if instance.handler_enum() == HandlerEnum.NORMAL:
            return self.__inner_do_normal_compute_execution(instance, params)
        else:
            return self.__inner_do_script_compute_execution(instance, dataComputeExecution, params)


    def _send_results(self,params, resultList,outputFdCodes):
        """
        发送结果数据
        :param params:
        :param resultList:
        :param outputFdCodes:
        :return:
        """
        if resultList is None or len(resultList) == 0:
            # 搞完了,没的结果,不处理
            return SUCCESS
        # 有结果,就需要从crawl的配置中找到目标的fd,然后调用fd的接口进行保存

        try:
            self.send_output_result(params["code_"], resultList, outputFdCodes)
        except Exception as e:
            FrameworkHandlerLogger.log_error(
                f"{params['site_']}_{params['projectCode_']}_{params['code_']} 发送结果数据失败,",e)
        return SUCCESS

    def __inner_do_script_crawl(self, instance, dataCrawl, params):
        simpleScriptCrawl = instance
        self.current_task = simpleScriptCrawl

        # 赋值 日志的上下文
        from schedule.utils.log.TaskLogger import TaskLogger
        from schedule.utils.log import TaskLogFileAppender
        simpleScriptCrawl.task_logger = TaskLogger(
            TaskLogFileAppender.get_log_file_path(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                                  params["__logId__"]))

        try:
            self.set_local_code(
                f"{dataCrawl.site}_{dataCrawl.projectCode}_{dataCrawl.code}")
            self.set_process_id(params.get("__processLogId__"))
            ScriptHandlerHolder.holder.key = self
            resultList = simpleScriptCrawl.do_crawl(params)
        except Exception as e:
            FrameworkHandlerLogger.log_error(
                f"{dataCrawl.site}_{dataCrawl.projectCode}_{dataCrawl.code} 执行失败,", e)
            import traceback
            return ResultVO(FAIL_CODE,
                            f"{dataCrawl.site}_{dataCrawl.projectCode}_{dataCrawl.code} 执行失败,\n{traceback.format_exc()}")
        finally:
            self.clear_local_code()
            self.clear_process_id()
            ScriptHandlerHolder.holder.key = None
            self.current_task = None
            simpleScriptCrawl.task_logger = None

        return self._send_results(params, resultList, dataCrawl.outputFdCodes)

    def __inner_do_normal_crawl(self, instance, params):
        self.current_task = instance

        # TODO 这里可能需要补充一些 taskThread里面的东西
        if params is None:
            params = {}

        # 补充sharding的信息
        try:
            from schedule import ShardContext
            sharding = ShardContext.get_sharding()
            params["__sharding_index__"] = sharding.index
            params["__sharding_total__"] = sharding.total
        except:
            pass

        try:
            return instance.execute(params)
        finally:
            # 如果是普通的,就不能把current_task设置为空
            # self.current_task = None
            pass

    def __inner_do_script_compute_execution(self, instance, dataComputeExecution, params):
        simpleScriptCE = instance
        self.current_task = simpleScriptCE

        # 赋值 日志的上下文
        from schedule.utils.log.TaskLogger import TaskLogger
        from schedule.utils.log import TaskLogFileAppender
        simpleScriptCE.task_logger = TaskLogger(
            TaskLogFileAppender.get_log_file_path(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                                  params["__logId__"]))

        try:
            self.set_local_code(
                f"{dataComputeExecution.site}_{dataComputeExecution.projectCode}_{dataComputeExecution.code}")
            self.set_process_id(params.get("__processLogId__"))
            ScriptHandlerHolder.holder.key = self
            resultList = simpleScriptCE.do_compute(FdInputPipe(dataComputeExecution.sourceFdCodes), params)
        except Exception as e:
            FrameworkHandlerLogger.log_error(
                f"{dataComputeExecution.site}_{dataComputeExecution.projectCode}_{dataComputeExecution.code} 执行失败,", e)
            import traceback
            return ResultVO(FAIL_CODE,
                            f"{dataComputeExecution.site}_{dataComputeExecution.projectCode}_{dataComputeExecution.code} 执行失败,\n{traceback.format_exc()}")
        finally:
            self.clear_local_code()
            self.clear_process_id()
            ScriptHandlerHolder.holder.key = None
            self.current_task = None
            simpleScriptCE.task_logger = None

        return self._send_results(params, resultList, dataComputeExecution.outputFdCodes)

    def __inner_do_normal_compute_execution(self, instance, params):
        self.current_task = instance

        # TODO 这里可能需要补充一些 taskThread里面的东西
        try:
            if params is None:
                params = {}

            #补充sharding的信息
            try:
                from schedule import ShardContext
                sharding = ShardContext.get_sharding()
                params["__sharding_index__"] = sharding.index
                params["__sharding_total__"] = sharding.total
            except:
                pass

            return instance.execute(params)
        finally:
            # 如果是普通的,就不能把current_task设置为空
            # self.current_task = None
            pass



#!/usr/bin/env python
from py_eureka_client import eureka_client

from api.exceptions import ServiceNoInstanceException, IllegalArgumentException
from client.ClientAuth import ClientAuth

FICUS_APP_NAME = "sobeyficus"

def do_service(service="", return_type="json",app_name=FICUS_APP_NAME,
               prefer_ip=False, prefer_https=False,
               method="GET", headers=None,params = None,
               data=None, timeout=None, auth=True):

    def walk_using_requests(url):

        global r
        try:
            if method.lower() == "get":
                r = requests.get(url, json=data, auth=ClientAuth() if auth else None,headers=headers,timeout=timeout,params=params)
            elif method.lower() =="post":
                r = requests.post(url, json=data, auth=ClientAuth() if auth else None,headers=headers,timeout=timeout,params=params)
            elif method.lower() == "put":
                r = requests.put(url, json=data, auth=ClientAuth() if auth else None,headers=headers,timeout=timeout,params=params)
            elif method.lower() =="patch":
                r = requests.patch(url, json=data, auth=ClientAuth() if auth else None,headers=headers,timeout=timeout,params=params)
            elif method.lower() =="delete":
                r = requests.delete(url, json=data, auth=ClientAuth() if auth else None,headers=headers,timeout=timeout,params=params)
            else:
                raise IllegalArgumentException(f"不支持的Rest操作:{method}")
            r.raise_for_status()
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
            from urllib.error import HTTPError
            raise HTTPError(url,500,"连接错误或超时",str(e),None)

        try:
            if return_type is None or return_type.lower() == "none":
                return None

            if len(r.content) <= 0:
                return None

            if return_type.lower() in ("json", "dict", "dictionary"):
                r.encoding = 'utf-8'
                return r.json()
            else:
                r.encoding = 'utf-8'
                return r.text
        finally:
            r.close()

    cli = eureka_client.get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")

    return cli.walk_nodes(app_name, service, prefer_ip, prefer_https, walk_using_requests)

def check_instance_avaliable(app_name=FICUS_APP_NAME):
    cli = eureka_client.get_discovery_client()
    if cli is None:
        raise Exception("Discovery Client has not initialized. ")
    app = cli.applications.get_application(app_name.upper())
    if app.instances is None or len(app.instances)==0:
        raise ServiceNoInstanceException(f"{app_name}服务没有找到可用的实例")

from .ComputeExecutionClient import *
from .DataCrawlClient import *
from .HandlerLogClient import *
from .JobScheduleClient import *
from .ScheduleCacheClient import *
from .ScheduleJobTaskLogClient import *



from typing import Optional

from api.handler.script import ScriptPythonFactory
from api.model.MetaModelTransform import MetaModelTransform
from factdatasource.transformer.ChineseToNumberTransformer import ChineseToNumberTransformer
from factdatasource.transformer.CompleteGB32100Transformer import CompleteGB32100Transformer
from factdatasource.transformer.CompleteStringTransformer import CompleteStringTransformer
from factdatasource.transformer.DefaultValueTransformer import DefaultValueTransformer
from factdatasource.transformer.FixItemTransformer import FixItemTransformer
from factdatasource.transformer.ITransformer import ITransformer
from factdatasource.transformer.NumberToChineseTransformer import NumberToChineseTransformer
from factdatasource.transformer.ParseDateTimeTransformer import ParseDateTimeTransformer

__instance = {
    "DefaultValue": DefaultValueTransformer(),
    "FixItem": FixItemTransformer(),
    "CompleteString": CompleteStringTransformer(),
    "ParseDate": ParseDateTimeTransformer(),
    "ChineseToNumber": ChineseToNumberTransformer(),
    "NumberToChinese": NumberToChineseTransformer(),
    "CompleteGB32100": CompleteGB32100Transformer(),
    "CustomTransformer": None
}


def get_transformer(transformer: MetaModelTransform) -> Optional[ITransformer]:
    transformer_name = transformer.transformer

    if transformer_name not in __instance:
        return None

    if "CustomTransformer" == transformer_name:
        # TODO 如果是自定义的转换器,那么就需要实例化
        source_code = transformer.params["__sourceCode__"]
        key = f"{str(transformer)}_{transformer.timeString()}"
        for name, instance in __instance.items():
            if name.startswith(str(transformer)):
                # 说明是同一个,然后就看时间是否是一样的
                if name == key:
                    # 时间都是一样的,说明是同一个,直接返回
                    return instance
                else:
                    # 说明是同一个转换器,但时间不同,需要重新加载
                    ScriptPythonFactory.destroy_instance(key)

        instance = ScriptPythonFactory.load_instance(key, source_code, "ITransformer")
        __instance[key] = instance
        return instance
    else:
        return __instance[transformer_name]



from flask import Blueprint
from init import app

remote = Blueprint('remote', __name__, )


@remote.errorhandler(Exception)
def default_error_handler(error):
    app.log.exception('error 500: %s', error)
    import traceback
    trac = traceback.format_exc()
    response = {"status": False, "message": str(error), "track": trac}

    return jsonify(response), 500, {'Content-Type': 'application/json'}

from .ActuatorRemote import *
from .TriggerRemoteActor import *
from .LogReadRemoteService import *
from .AsyncResponseRestService import *
from .PrometheusMetricsRemoteService import *



from api.model.GraphNode import GraphNode


class GraphRelation(dict):

    """
        图数据的关系对象
        """

    def __init__(self, from_: GraphNode,to:GraphNode, type: str,guid:str, tags: dict = None) -> None:
        """
        图关系节点
        :param from_: 关系开始节点唯一guid
        :param to: 关系结束节点唯一guid
        :param type: 关系类型
        :param guid: 关系的唯一guid
        :param tags: 辅助标签/属性
        """
        super().__init__()
        self.__setitem__("@type", "com.sobey.jcg.sobeycube.core.api.domain.graph.GraphRelation")
        self.__setitem__("from", from_)
        self.__setitem__("to", to)
        self.__setitem__("guid", guid)
        self.__setitem__("type", type)
        if tags is not None:
            self.__setitem__("tags", tags)

    @property
    def from_(self):
        return self.__getitem__("from")

    @from_.setter
    def from_(self, value: GraphNode):
        self.__setitem__("from", value)

    @property
    def to(self):
        return self.__getitem__("to")

    @to.setter
    def to(self, value: str):
        self.__setitem__("to", value)

    @property
    def guid(self):
        return self.__getitem__("guid")

    @guid.setter
    def guid(self, value: str):
        self.__setitem__("guid", value)

    @property
    def type(self):
        return self.__getitem__("type")

    @type.setter
    def type(self, value: str):
        self.__setitem__("type", value)

    @property
    def tags(self):
        return self.__getitem__("tags") if "tags" in self else None

    @tags.setter
    def tags(self, value: dict):
        if value is not None:
            self.__setitem__("tags", value)
        else:
            self.__delitem__("tags")


from abc import abstractmethod

from api.handler import HandlerEnum
from api.handler.IHandler import IHandler
from api.handler.script.ICacheAbleScript import ICacheAbleScript
from api.handler.IKillable import IKillable
from api.handler.ILogAbleHandler import ILogAbleHandler
from api.handler.script.IProgressAbleScript import IProgressAbleScript
from api.model.FdInputPipe import FdInputPipe


class ISimpleScriptCE(IKillable, ILogAbleHandler, ICacheAbleScript, IHandler, IProgressAbleScript):
    """
    Python脚本式的CE的基类
    """

    @abstractmethod
    def do_compute(self, source_fds: FdInputPipe, params: dict):
        """
        计算器的业务逻辑
        :param source_fds: 输入源FD
        :param params: 计算器上下文参数
        :return: 返回一个 OutputWrapper的 数组
        """
        pass

    def handler_enum(self) -> HandlerEnum:
        return HandlerEnum.SCRIPT

    def is_killed(self) -> bool:
        """
        判断是否杀掉了
        :return:
        """
        return False



"""
认证模块操作
"""
import requests
from munch import Munch

from api.exceptions import IllegalArgumentException
from client import do_service, check_instance_avaliable


def get_authorization_header(client_id, client_secret):
    import base64
    return 'Basic ' + str(base64.b64encode(f'{client_id}:{client_secret}'.encode('utf-8')), encoding='utf-8')


def oauth_access_token(server, grant_type, client_id, client_secret, username, password):
    """
    获取oauth系统认证的accessToken
    :param client_id:
    :param client_secret:
    :return:
    """
    check_instance_avaliable()

    try:
        from urllib.parse import quote
        # url 中需要编码处理下，避免其中有特殊字符
        url = f"{server}/oauth/token?grant_type={quote(grant_type)}&scope=all&username={quote(str(username))}&password={quote(str(password))}"
        authorization = get_authorization_header(client_id, client_secret)

        r = requests.post(url, headers={"Authorization": authorization})

        if r is not None and r.status_code == 200:
            entity = Munch(r.json())

            from datetime import datetime
            import time
            return entity["access_token"], datetime.fromtimestamp(time.time() + entity["expires_in"])
        else:
            return None, None
    except requests.exceptions.HTTPError as ee:
        raise IllegalArgumentException(f"从ficus获取oauth签名失败,{ee.response.status_code}")


def oauth_jwt_token(user_name, user_password):
    """
    获取jwt的用户内容权限的jwtToken
    :param user_name:
    :param user_password:
    :return:
    """

    check_instance_avaliable()

    try:
        r = do_service("/user/login", method="POST", return_type="str",
                       data={"username": user_name, "password": user_password}, auth=None)

        if r is not None:
            jwt_token = r
            return (jwt_token, _find_expiration(jwt_token))
        else:
            return (None, None)
    except requests.exceptions.HTTPError as ee:
        raise IllegalArgumentException(f"从ficus获取用户认证失败,{ee.response.status_code}")


def _find_expiration(jwt_token: str):
    """
    解析jwtToken的过期时间
    :param jwt_token:
    :return:
    """

    # jwt有三部分,每个部分用逗号分隔.
    if jwt_token is None:
        return None

    splits = jwt_token.split(".")

    # 取第二部分
    s = splits[1]

    # 把第二部分用base64解码.
    import base64
    s1 = base64.b64decode(s)

    # 出来是一个map,取exp字段
    import json
    json_object = json.loads(s1)

    exp = json_object["exp"]

    from datetime import datetime
    return datetime.fromtimestamp(exp)
